# -*- coding: utf-8 -*-
"""This module loads settings from a configuration file."""
__author__ = "ajshajib"

import sys
from lenstronomy.Workflow.fitting_sequence import FittingSequence
from schwimmbad import choose_pool

from .files import FileSystem
from .config import ModelConfig
from .data import ImageData
from .data import PSFData
from .recipe import Recipe


class Processor(object):
    """This class contains methods to model a single lens system or a bunch of systems
    from the config files."""

    def __init__(self, io_directory):
        """

        :param io_directory: path to the input/output directory. Should not
            end with slash.
        :type io_directory: `str`
        """
        self.io_directory = io_directory
        self.file_system = FileSystem(io_directory)
        self.lens_list = self.file_system.get_lens_list()

    def swim(
        self,
        lens_name,
        model_id,
        log=True,
        mpi=False,
        recipe_name="galaxy-quasar",
        thread_count=1,
    ):
        """Run models for a single lens.

        :param lens_name: lens name
        :type lens_name: `str`
        :param model_id: identifier for the model run
        :type model_id: `str`
        :param log: if `True`, all `print` statements will be logged. This should be `False` if running in a notebook.
        :type log: `bool`
        :param mpi: MPI option
        :type mpi: `bool`
        :param recipe_name: recipe for pre-sampling optimization, supported
            ones now: 'galaxy-quasar' and 'galaxy-galaxy'
        :type recipe_name: `str`
        :param sampler: 'EMCEE' or 'COSMOHAMMER', cosmohammer is kept for
            legacy
        :type sampler: `str`
        :param thread_count: number of threads if `multiprocess` is used
        :type thread_count: `int`
        :return:
        :rtype:
        """
        pool = choose_pool(mpi=mpi)

        if log and pool.is_master():
            log_file = open(
                self.file_system.get_log_file_path(lens_name, model_id), "wt"
            )
            sys.stdout = log_file

        config = self.get_lens_config(lens_name)
        recipe = Recipe(config, thread_count=thread_count)

        psf_supersampling_factor = config.get_psf_supersampled_factor()
        kwargs_data_joint = self.get_kwargs_data_joint(
            lens_name, psf_supersampled_factor=psf_supersampling_factor
        )

        fitting_sequence = FittingSequence(
            kwargs_data_joint,
            config.get_kwargs_model(),
            config.get_kwargs_constraints(),
            config.get_kwargs_likelihood(),
            config.get_kwargs_params(),
            mpi=mpi,
        )

        fitting_kwargs_list = recipe.get_recipe(
            kwargs_data_joint=kwargs_data_joint, recipe_name=recipe_name
        )
        print(f"Optimizing model for {lens_name} with recipe: {recipe_name}.")

        fit_output = fitting_sequence.fit_sequence(fitting_kwargs_list)
        kwargs_result = fitting_sequence.best_fit(bijective=False)
        multi_band_list_out = fitting_sequence.multi_band_list

        output = {
            "settings": config.settings,
            "kwargs_result": kwargs_result,
            "fit_output": fit_output,
            "multi_band_list_out": multi_band_list_out,
        }

        if pool.is_master():
            self.file_system.save_output(lens_name, model_id, output)

        if log and pool.is_master():
            log_file.close()

    def get_lens_config(self, lens_name):
        """Get the `ModelConfig` object for a lens.

        :param lens_name: lens name
        :type lens_name: `str`
        :return: `ModelConfig` instance
        :rtype:
        """
        return ModelConfig(lens_name, file_system=self.file_system)

    def get_kwargs_data_joint(self, lens_name, psf_supersampled_factor=1):
        """Create `kwargs_data` for a lens and given filters.

        :param lens_name: lens name
        :type lens_name: `str`
        :param psf_supersampled_factor: Supersampled factor of given PSF.
        :rtype psf_supersampled_factor: `float`
        :return:
        :rtype:
        """
        config = self.get_lens_config(lens_name)

        bands = config.settings["band"]

        kwargs_numerics = config.get_kwargs_numerics()

        multi_band_list = []

        for b, kwargs_num in zip(bands, kwargs_numerics):
            image_data = self.get_image_data(lens_name, b)
            psf_data = self.get_psf_data(lens_name, b)

            psf_data.kwargs_psf["point_source_supersampling_factor"] = (
                psf_supersampled_factor
            )

            multi_band_list.append(
                [image_data.kwargs_data, psf_data.kwargs_psf, kwargs_num]
            )

        kwargs_data_joint = {
            "multi_band_list": multi_band_list,
            "multi_band_type": "multi-linear",
        }

        return kwargs_data_joint

    def get_image_data(self, lens_name, band):
        """Get the `ImageData` instance.

        :param lens_name: name of the lens
        :type lens_name: `str`
        :param band: image band/filter
        :type band: `str`
        :return: `ImageData` instance
        :rtype:
        """
        return ImageData(self.file_system.get_image_file_path(lens_name, band))

    def get_psf_data(self, lens_name, band):
        """Get the `PSFData` instance.

        :param lens_name: lens name
        :type lens_name: `str`
        :param band: image band/filter
        :type band: `str`
        :return: `PSFData` instance
        :rtype:
        """
        return PSFData(self.file_system.get_psf_file_path(lens_name, band))



# -*- coding: utf-8 -*-
"""This module loads data and psfs from data files."""
__author__ = "ajshajib"

import h5py
import numpy as np
from copy import deepcopy
from lenstronomy.Data.coord_transforms import Coordinates


class Data(object):
    """This is a superclass to load datafiles."""

    def __init__(self):
        pass

    @staticmethod
    def load_from_file(file_path):
        """Load data from h5py file.

        :param file_path: path to a file
        :type file_path: `str`
        :return:
        :rtype:
        """
        with h5py.File(file_path, "r") as f:
            data = {}
            for key in f:
                data[key] = f[key][()]

        return data


class ImageData(Data):
    """This class contains the image of a lens system."""

    def __init__(self, data_file_path):
        """

        :param data_file_path: path to a data file
        :type data_file_path: `str`
        """
        super().__init__()

        self._data = self.load_from_file(data_file_path)

    @property
    def kwargs_data(self):
        """Get `kwargs_data` dictionary.

        :return: `kwargs_data`
        :rtype: `dict`
        """
        kwargs_data = deepcopy(self._data)

        return kwargs_data

    def get_image(self):
        """Get image `ndarray` from the saved in the class instance.

        :return: image
        :rtype: `ndarray`
        """
        return deepcopy(self._data["image_data"])

    def get_image_coordinate_system(self):
        """Get the coordinate system of the image data.

        :return: coordinate system
        :rtype: `str`
        """
        ra_at_xy_0 = self.kwargs_data["ra_at_xy_0"]
        dec_at_xy_0 = self.kwargs_data["dec_at_xy_0"]
        transform_pix2angle = np.array(self.kwargs_data["transform_pix2angle"])

        coordinate_system = Coordinates(transform_pix2angle, ra_at_xy_0, dec_at_xy_0)

        return coordinate_system

    def get_image_size(self):
        """Get the number of pixels in the image.

        :return: number of pixels
        :rtype: `int`
        """
        return self.kwargs_data["image_data"].shape[0]

    def get_image_pixel_scale(self):
        """Get the pixel scale of the image.

        :return: pixel scale
        :rtype: `float`
        """
        return self.get_image_coordinate_system().pixel_width


class PSFData(Data):
    """This class contains the PSF for a lens system."""

    def __init__(self, psf_file_path):
        """

        :param psf_file_path: path to a PSF data file
        :type psf_file_path: `str`
        """
        super().__init__()

        self._data = self.load_from_file(psf_file_path)

    @property
    def kwargs_psf(self):
        """Get `kwargs_psf` dictionary.

        :return: `kwargs_psf`
        :rtype: `dict`
        """
        kwargs_psf = deepcopy(self._data)
        kwargs_psf["psf_type"] = "PIXEL"
        kwargs_psf["kernel_point_source_init"] = deepcopy(
            kwargs_psf["kernel_point_source"]
        )
        return kwargs_psf



# -*- coding: utf-8 -*-
from ..processor.data import ImageData
from ..processor.files import FileSystem


class AI(object):
    """This parent class holds common methods for AI classes."""

    def __init__(self, io_directory_path):
        """

        :param io_directory_path: path to the input-output directory
        :type io_directory_path: `str`
        """
        self.file_system = FileSystem(io_directory_path)

    def get_image_data(self, lens_name, band_name):
        """Get the image data.

        :param lens_name: lens name
        :type lens_name: `str`
        :param band_name: band name
        :type band_name: `str`
        """
        image_path = self.file_system.get_image_file_path(lens_name, band_name)
        image_data = ImageData(image_path)

        return image_data



# -*- coding: utf-8 -*-
"""Tests for config module."""
from pathlib import Path
from dolphin.processor import Processor
from dolphin.ai import Vision
from dolphin.ai import Modeler


class TestPipeline(object):
    """Test pipeline for the AI module."""

    def setup_method(self):
        """Setup method for the test class."""

        _ROOT_DIR = Path(__file__).resolve().parents[1]
        self.io_directory_path = str(_ROOT_DIR / "io_directory_example")

    def test_quasar_pipeline(self):
        """Test the entire pipeline."""
        lens_name = "lensed_quasar_2"
        band = "F814W"

        vision = Vision(io_directory_path=self.io_directory_path, source_type="quasar")
        vision.create_segmentation_for_single_lens(lens_name, band)

        modeler = Modeler(
            io_directory_path=self.io_directory_path, source_type="quasar"
        )
        modeler.create_config_for_single_lens(
            lens_name,
            band,
            psf_iteration_settings={
                "stacking_method": "median",
                "num_iter": 2,
                "psf_iter_factor": 0.5,
                "keep_psf_variance_map": True,
                "psf_symmetry": 4,
            },
            pso_settings={"num_particle": 4, "num_iteration": 2},
            sampler_settings=None,
        )

        processor = Processor(self.io_directory_path)
        processor.swim(
            lens_name=lens_name,
            model_id="example",
            log=True,
            recipe_name="galaxy-quasar",
        )

    def test_galaxy_pipeline(self):
        """Test the entire pipeline."""
        lens_name = "lens_system2"
        band = "F390W"

        vision = Vision(io_directory_path=self.io_directory_path, source_type="galaxy")
        vision.create_segmentation_for_single_lens(lens_name, band)

        modeler = Modeler(
            io_directory_path=self.io_directory_path, source_type="galaxy"
        )
        modeler.create_config_for_single_lens(
            lens_name,
            band,
            psf_iteration_settings=None,
            pso_settings={"num_particle": 4, "num_iteration": 2},
            sampler_settings=None,
        )

        processor = Processor(self.io_directory_path)
        processor.swim(
            lens_name=lens_name,
            model_id="example",
            log=False,
            recipe_name="galaxy-galaxy",
        )



# -*- coding: utf-8 -*-

# This module creates a configuration file from the output of the visual recognition model.


import numpy as np
import yaml

from .ai import AI


class Modeler(AI):
    """This class creates a configuration file from the output of the visual recognition
    model."""

    def __init__(self, io_directory_path, source_type):
        """Initialize the Configure object.

        :param io_directory_path: path to the input/output directory
        :type io_directory_path: `str`
        :param source_type: source type
        :type source_type: `str`
        """
        super().__init__(io_directory_path)
        self._source_type = source_type

    def create_configuration_for_all_lenses(self, band_name, **kwargs):
        """Create configuration files for all lenses.

        :param band_name: band name
        :type band_name: `str`
        :param kwargs: additional keyword arguments to be passed to `get_configuration`, check the documentation of that function.
        :type kwargs: `dict`
        """
        lens_list = self.file_system.get_lens_list()

        for lens_name in lens_list:
            self.create_config_for_single_lens(lens_name, band_name, **kwargs)

        print(f"Done creating configuration files for {len(lens_list)} lenses.")

    def create_config_for_single_lens(self, lens_name, band_name, **kwargs):
        """Create configuration file for a single lens.

        :param lens_name: lens name
        :type lens_name: `str`
        :param band_name: band name
        :type band_name: `str`
        :param kwargs: additional keyword arguments to be passed to `get_configuration`, check the documentation of that function.
        :type kwargs: `dict`
        :return: configuration
        :rtype: `dict`
        """
        configuration = self.get_configuration(
            lens_name,
            band_name,
            **kwargs,
        )
        self.save_configuration(configuration, lens_name)

        return configuration

    def load_semantic_segmentation(self, lens_name, band_name):
        """Load semantic segmentation output from the visual recognition model.

        :param lens_name: lens name
        :type lens_name: `str`
        :param band_name: band name
        :type band_name: `str`
        """
        return self.file_system.load_semantic_segmentation(lens_name, band_name)

    def save_configuration(self, config, lens_name):
        """Save the configuration to a YAML file.

        :param config: configuration
        :type config: `dict`
        :param lens_name: lens name
        :type lens_name: `str`
        """
        config_file_path = self.file_system.get_config_file_path(lens_name)

        with open(config_file_path, "w") as config_file:
            yaml.dump(config, config_file, default_flow_style=False, sort_keys=False)

    def get_configuration(
        self,
        lens_name,
        band_name,
        pso_settings={"num_particle": 50, "num_iteration": 50},
        psf_iteration_settings=None,
        sampler_name="emcee",
        sampler_settings=None,
        supersampling_factor=[3],
        max_satellite_number=2,
        minimum_satellite_area=15,
        satellite_bound=0.25,
        clear_center=0.2,
        source_n_max=6,
        mask_radius_factor=2.5,
    ):
        """Get configuration from the semantic segmentation output. This method
        currently works only for the single-band case.

        :param lens_name: lens name
        :type lens_name: `str`
        :param band_name: band name
        :type band_name: `str`
        :param pso_settings: PSO settings
        :type pso_settings: `dict`
        :param psf_iteration_settings: PSF iteration settings. If `None`, PSF
            iteration will not be performed. If settings are provided, PSF
            iteration will be performed. Example settings:
            {
                "stacking_method": "median",
                "num_iter": 20,
                "psf_iter_factor": 0.5,
                "keep_psf_variance_map": True,
                "psf_symmetry": 4,
            }
        :type psf_iteration_settings: `dict`
        :param sampler_name: sampler name, default is "emcee"
        :type sampler_name: `str`
        :param sampler_settings: sampler settings. If `None`, sampling will not be
            performed. If settings are provided, sampling will be performed.
            Example settings:
            {
                "n_burn": 0,
                "n_run": 100,
                "walkerRatio": 2,
            }
        :type sampler_settings: `dict`
        :param supersampling_factor: supersampling factor
        :type supersampling_factor: `List[int]`
        :param max_satellite_number: maximum number of satellites
        :type max_satellite_number: `int`
        :param minimum_satellite_area: minimum satellite area
        :type minimum_satellite_area: `int`
        :param satellite_bound: satellite bound
        :type satellite_bound: `float`
        :param clear_center: radius (arcsecond) to clear the center from any detected satellite or quasar
        :type clear_center: `float`
        :param source_n_max: maximum number of shapelet coefficients for the source, choose `None` to turn off shapelets
        :type source_n_max: `int`
        :param mask_radius_factor: factor of initial Einstein radius estimate to set the circular mask radius
        :type mask_radius_factor: `float`
        :return: configuration
        :rtype: `dict`
        """
        # Get image data and coordinate system
        image_data = self.get_image_data(lens_name, band_name)
        coordinate_system = image_data.get_image_coordinate_system()
        semantic_segmentation = self.load_semantic_segmentation(lens_name, band_name)

        # Initialize configuration dictionary
        config = {}
        config["lens_name"] = lens_name
        config["band"] = [band_name]

        # Define model components
        config["model"] = {
            "lens": ["EPL", "SHEAR_GAMMA_PSI"],
            "lens_light": ["SERSIC_ELLIPSE"],
            "source_light": ["SERSIC_ELLIPSE"],
            "point_source": (
                ["LENSED_POSITION"] if self._source_type == "quasar" else []
            ),
        }
        if source_n_max is not None:
            config["model"]["source_light"].append("SHAPELETS")

        # Set lens options
        galaxy_center_x, galaxy_center_y = self.get_lens_galaxy_center_init(
            semantic_segmentation, coordinate_system
        )
        config["lens_option"] = {
            "centroid_init": [galaxy_center_x.item(), galaxy_center_y.item()],
            "centroid_bound": 0.2,
        }
        config["lens_light_option"] = {"fix": {0: {"n_sersic": 4.0}}}
        if source_n_max is not None:
            config["source_light_option"] = {"n_max": [source_n_max]}

        # Set point source options
        if self._source_type == "quasar":
            point_source_init = self.get_quasar_image_position(
                semantic_segmentation, coordinate_system, clear_center=clear_center
            )
            config["point_source_option"] = {
                "ra_init": point_source_init[0].tolist(),
                "dec_init": point_source_init[1].tolist(),
                "bound": 0.2,
            }

        # Set satellite options
        satellite_positions = self.get_satellite_positions(
            semantic_segmentation,
            coordinate_system,
            clear_center=clear_center,
            minimum_pixel_area=minimum_satellite_area,
        )

        config["satellites"] = {
            "centroid_init": [a for a in satellite_positions[:max_satellite_number]],
            "centroid_bound": satellite_bound,
        }

        # Set guess params
        theta_E_init = self.get_theta_E_init(semantic_segmentation, coordinate_system)
        config["guess_params"] = {"lens": {0: {"theta_E": float(theta_E_init)}}}

        # Set numeric options
        config["numeric_option"] = {"supersampling_factor": supersampling_factor}

        # Set fitting options
        config["fitting"] = {
            "pso": pso_settings is not None,
            "pso_settings": pso_settings,
        }

        config["fitting"]["psf_iteration"] = (
            True
            if psf_iteration_settings is not None and self._source_type == "quasar"
            else False
        )
        config["fitting"]["psf_iteration_settings"] = psf_iteration_settings

        if config["fitting"]["psf_iteration"]:
            # Calculate distances between quasar images
            distances = []
            for i in range(len(point_source_init[0])):
                for j in range(i + 1, len(point_source_init[0])):
                    distances.append(
                        np.sqrt(
                            (point_source_init[0][i] - point_source_init[0][j]) ** 2
                            + (point_source_init[1][i] - point_source_init[1][j]) ** 2
                        )
                    )

            # Set PSF iteration settings
            # Second minimum distance is used to set the block center neighbour and error map radius
            if (
                "block_center_neighbour"
                not in config["fitting"]["psf_iteration_settings"]
            ):
                config["fitting"]["psf_iteration_settings"][
                    "block_center_neighbour"
                ] = float(np.sort(distances)[1] / 2.0)
            if "error_map_radius" not in config["fitting"]["psf_iteration_settings"]:
                config["fitting"]["psf_iteration_settings"]["error_map_radius"] = float(
                    np.sort(distances)[1] / 2.0
                )

        # Set sampling options
        config["fitting"]["sampling"] = sampler_settings is not None
        config["fitting"]["sampler"] = sampler_name
        config["fitting"]["sampler_settings"] = sampler_settings

        # Set mask options
        if self._source_type == "quasar":
            config["mask"] = {}
            config["mask"]["provided"] = True
            self.file_system.save_mask(
                lens_name,
                band_name,
                self.get_mask_from_semantic_segmentation(
                    semantic_segmentation,
                    coordinate_system,
                    mask_radius_factor=mask_radius_factor,
                ),
            )

        return config

    def get_mask_from_semantic_segmentation(
        self, semantic_segmentation, coordinate_system, mask_radius_factor=2.5
    ):
        """Get mask from the semantic segmentation output.

        :param semantic_segmentation: semantic segmentation output
        :type semantic_segmentation: `numpy.ndarray`
        :param coordinate_system: coordinate system
        :type coordinate_system: `Coordinates`
        :param image_positions: image positions
        :type image_positions: `List[np.ndarray]`
        :param mask_radius_factor: factor of initial Einstein radius estimate to set the circular mask radius
        :type mask_radius_factor: `float`
        :return: mask
        :rtype: `numpy.ndarray`
        """
        theta_E_init = self.get_theta_E_init(semantic_segmentation, coordinate_system)
        galaxy_center_x, galaxy_center_y = self.get_lens_galaxy_center_init(
            semantic_segmentation, coordinate_system
        )

        mask = np.zeros(semantic_segmentation.shape)

        galaxy_center_x_pixel, galaxy_center_y_pixel = coordinate_system.map_coord2pix(
            galaxy_center_x, galaxy_center_y
        )
        pixel_size = coordinate_system.pixel_width
        for i in range(mask.shape[0]):
            for j in range(mask.shape[1]):
                if (j - galaxy_center_x_pixel) ** 2 + (
                    i - galaxy_center_y_pixel
                ) ** 2 < (mask_radius_factor * theta_E_init / pixel_size) ** 2:
                    mask[i, j] = 1

        return mask

    def get_theta_E_init(self, semantic_segmentation, coordinate_system):
        """Get the initial guess for the Einstein radius.

        :param galaxy_center: galaxy center
        :type galaxy_center: `List[float]`
        :param image_positions: image positions
        :type image_positions: `List[List[float]]`
        :return: initial guess for the Einstein radius
        :rtype: `float`
        """
        galaxy_center = self.get_lens_galaxy_center_init(
            semantic_segmentation, coordinate_system
        )

        if self._source_type == "quasar":
            image_positions = self.get_quasar_image_position(
                semantic_segmentation, coordinate_system
            )
            x_0, y_0 = galaxy_center
            x_i, y_i = image_positions
            theta_E_init = np.mean(np.sqrt((x_i - x_0) ** 2 + (y_i - y_0) ** 2))
        elif self._source_type == "galaxy":
            # Get all x and y indices of the pixels where semantic segmentation is 2
            arc_x, arc_y = np.where(semantic_segmentation == 2)

            arc_ra, arc_dec = coordinate_system.map_pix2coord(arc_y, arc_x)
            rs = np.sqrt(
                (arc_ra - galaxy_center[0]) ** 2 + (arc_dec - galaxy_center[1]) ** 2
            )

            theta_E_init = np.mean(rs)

        return theta_E_init

    @classmethod
    def get_lens_galaxy_center_init(cls, semantic_segmentation, coordinate_system):
        """Get the initial guess for the lens galaxy center.

        :param semantic_segmentation: semantic segmentation output
        :type semantic_segmentation: `numpy.ndarray`
        :param coordinate_system: coordinate system
        :type coordinate_system: `Coordinates`
        :return: initial guess for the lens galaxy center
        :rtype: `[float, float]`
        """
        galaxy_center_pixels = cls.list_region_centers(semantic_segmentation, 1)
        galaxy_center_ra, galaxy_center_dec = coordinate_system.map_pix2coord(
            galaxy_center_pixels[0][1], galaxy_center_pixels[0][0]
        )
        return [galaxy_center_ra, galaxy_center_dec]

    def get_quasar_image_position(
        self, semantic_segmentation, coordinate_system, clear_center=0.0
    ):
        """Identify quasar image positions from the semantic segmentation output.

        :param semantic_segmentation: semantic segmentation output
        :type semantic_segmentation: `numpy.ndarray`
        :param coordinate_system: coordinate system
        :type coordinate_system: `Coordinates`
        :param clear_center: radius (arcsecond) to clear the center from any detected quasar
        :type clear_center: `float`
        :return: quasar image positions
        :rtype: `[np.ndarray, np.ndarray]`
        """
        if self._source_type != "quasar":
            raise NotImplementedError(
                "Quasar image position calculation is only implemented for quasar sources."
            )
        quasar_positions = self.list_region_centers(semantic_segmentation, 3)

        quasar_ra, quasar_dec = [], []

        galaxy_center = self.get_lens_galaxy_center_init(
            semantic_segmentation, coordinate_system
        )

        for position in quasar_positions:
            ra, dec = coordinate_system.map_pix2coord(position[1], position[0])

            distance = np.sqrt(
                (ra - galaxy_center[0]) ** 2 + (dec - galaxy_center[1]) ** 2
            )

            if distance > clear_center:
                quasar_ra.append(ra)
                quasar_dec.append(dec)

        return [np.array(quasar_ra), np.array(quasar_dec)]

    @classmethod
    def get_satellite_positions(
        cls,
        semantic_segmentation,
        coordinate_system,
        clear_center=0.0,
        minimum_pixel_area=1,
    ):
        """Identify satellite positions from the semantic segmentation output.

        :param semantic_segmentation: semantic segmentation output
        :type semantic_segmentation: `numpy.ndarray`
        :param coordinate_system: coordinate system
        :type coordinate_system: `Coordinates`
        :param clear_center: radius (arcsecond) to clear the center from any detected satellite
        :type clear_center: `float`
        :param minimum_pixel_area: minimum pixel area for a satellite
        :type minimum_pixel_area: `int`
        :return: satellite positions
        :rtype: `List[List[float]]`
        """
        satellite_positions = cls.list_region_centers(
            semantic_segmentation, 4, minimum_pixel_area=minimum_pixel_area
        )
        galaxy_center = cls.get_lens_galaxy_center_init(
            semantic_segmentation, coordinate_system
        )

        satellite_ra, satellite_dec = [], []

        for position in satellite_positions:
            ra, dec = coordinate_system.map_pix2coord(position[1], position[0])

            distance = np.sqrt(
                (ra - galaxy_center[0]) ** 2 + (dec - galaxy_center[1]) ** 2
            )
            if distance > clear_center:
                satellite_ra.append(float(ra))
                satellite_dec.append(float(dec))

        return [[ra, dec] for ra, dec in zip(satellite_ra, satellite_dec)]

    @classmethod
    def collect_connected_pixels(
        cls, x, y, pixels, visited, target_value, matrix, rows, cols
    ):
        """Collect all connected pixels in the matrix.

        :param x: x-coordinate
        :type x: `int`
        :param y: y-coordinate
        :type y: `int`
        :param pixels: list of pixels
        :type pixels: `List[Tuple[int, int]]`
        :param visited: matrix to track visited pixels
        :type visited: `List[List[bool]]`
        :param target_value: target value to match
        :type target_value: `int`
        :param matrix: input matrix
        :type matrix: `List[List[int]]`
        :param rows: number of rows in the matrix
        :type rows: `int`
        :param cols: number of columns in the matrix
        :type cols: `int`
        """
        # Check if the current pixel is out of bounds
        if x < 0 or x >= rows or y < 0 or y >= cols:
            return

        # Check if the current pixel is already visited or does not match the target value
        if visited[x][y] or matrix[x][y] != target_value:
            return

        # Mark the current pixel as visited and add it to the list of pixels
        visited[x][y] = True
        pixels.append((x, y))

        # Explore all 4 possible directions (up, down, left, right)
        cls.collect_connected_pixels(
            x + 1, y, pixels, visited, target_value, matrix, rows, cols
        )
        cls.collect_connected_pixels(
            x - 1, y, pixels, visited, target_value, matrix, rows, cols
        )
        cls.collect_connected_pixels(
            x, y + 1, pixels, visited, target_value, matrix, rows, cols
        )
        cls.collect_connected_pixels(
            x, y - 1, pixels, visited, target_value, matrix, rows, cols
        )

    @classmethod
    def list_region_centers(cls, matrix, target_value, minimum_pixel_area=1):
        """List the central pixel (x, y) for all regions in a matrix with the target
        value.

        :param matrix: input matrix
        :type matrix: `List[List[int]]`
        :param target_value: target value
        :type target_value: `int`
        :return: list of central pixels (x, y) for all regions
        :rtype: `List[Tuple[int, int]]`
        :param minimum_pixel_area: minimum size of the region
        :type minimum_pixel_area: `int`
        :return: list of central pixels (x, y) for all regions
        :rtype: `List[Tuple[int, int]]`
        """
        rows, cols = len(matrix), len(matrix[0])
        visited = [[False for _ in range(cols)] for _ in range(rows)]
        region_centers = []
        region_areas = []

        for i in range(rows):
            for j in range(cols):
                if matrix[i][j] == target_value and not visited[i][j]:
                    pixels = []
                    cls.collect_connected_pixels(
                        i, j, pixels, visited, target_value, matrix, rows, cols
                    )
                    if pixels and len(pixels) >= minimum_pixel_area:
                        center_x = sum(p[0] for p in pixels) // len(pixels)
                        center_y = sum(p[1] for p in pixels) // len(pixels)

                        region_centers.append((center_x, center_y))
                        region_areas.append(len(pixels))

        # sort by region_areas
        region_centers = [
            x for _, x in sorted(zip(region_areas, region_centers), reverse=True)
        ]

        return region_centers



# -*- coding: utf-8 -*-
"""This module creates `fitting_kwargs_list` for `FittingSequence.fit_sequence()` with
pre-defined recipes."""
__author__ = "ajshajib"

from copy import deepcopy
import numpy as np
from scipy import ndimage


class Recipe(object):
    """This class contains methods to create fitting recipes.

    It builds an optimization workflow (currently using particle-swarm optimization) to
    first find a good enough lens model within the total parameter space. Then, the
    sampling can be done starting from the neighborhood of this point.
    """

    def __init__(self, config, thread_count=1):
        """Initiate the class from the given settings for a lens system.

        :param config: `ModelConfig` instance
        :type config: `class`
        :param thread_count: number of threads if `multiprocess` is used
        :type thread_count: `int`
        """
        self._config = config
        try:
            config.settings["fitting"]["pso"]
        except (NameError, KeyError):
            self.do_pso = False
        else:
            self.do_pso = deepcopy(config.settings["fitting"]["pso"])

            self._pso_num_particle = self._config.settings["fitting"]["pso_settings"][
                "num_particle"
            ]
            self._pso_num_iteration = self._config.settings["fitting"]["pso_settings"][
                "num_iteration"
            ]

            if self.do_pso is None:
                self.do_pso = False

        try:
            config.settings["fitting"]["psf_iteration"]
        except (NameError, KeyError):
            self.reconstruct_psf = False
        else:
            self.reconstruct_psf = deepcopy(config.settings["fitting"]["psf_iteration"])

            if self.reconstruct_psf is None:
                self.reconstruct_psf = False

        try:
            config.settings["fitting"]["sampling"]
        except (NameError, KeyError):
            self.do_sampling = False
        else:
            self.do_sampling = deepcopy(config.settings["fitting"]["sampling"])

            if self.do_sampling is None:
                self.do_sampling = False

        self._thread_count = thread_count

        self.guess_params = {}
        for component in ["lens", "source", "lens_light", "ps"]:
            try:
                self.guess_params[component] = deepcopy(
                    config.settings["guess_params"][component]
                )
            except (NameError, KeyError):
                self.guess_params[component] = None

    def get_recipe(self, kwargs_data_joint=None, recipe_name="galaxy-quasar"):
        """Get `fitting_kwargs_list` according to the requested `recipe`.

        :param config: `ModelConfig` instance
        :type config:
        :param kwargs_data_joint: `kwargs_data_joint` dictionary
        :type kwargs_data_joint:
        :param recipe_name: recipe name, 'galaxy-quasar' or 'galaxy-galaxy'
        :type recipe_name: `str`
        :return: fitting kwargs list
        :rtype: `list`
        """
        fitting_kwargs_list = []

        if recipe_name == "galaxy-quasar":
            try:
                # if this key exists then return it
                self._config.settings["fitting_kwargs_list"]
            except (KeyError, NameError):
                fitting_kwargs_list += self.get_galaxy_quasar_recipe()
            else:
                if self._config.settings["fitting_kwargs_list"] is not None:
                    fitting_kwargs_list += self._config.settings["fitting_kwargs_list"]
                else:
                    fitting_kwargs_list += self.get_galaxy_quasar_recipe()
        elif recipe_name == "galaxy-galaxy":
            if kwargs_data_joint is None:
                raise ValueError(
                    "kwargs_data_joint is necessary to use "
                    "galaxy-galaxy optimization recipe!"
                )
            fitting_kwargs_list += self.get_galaxy_galaxy_recipe(kwargs_data_joint)
        elif recipe_name == "skip":
            pass
        else:
            raise ValueError("Recipe name '{}' not recognized!!".format(recipe_name))

        fitting_kwargs_list += self.get_sampling_sequence()

        return fitting_kwargs_list

    def _get_power_law_model_index(self):
        """Get the index of the power-law model, if included in the lens model list.

        :return: index or `None`
        :rtype: `int`
        """
        lens_model_list = self._config.get_lens_model_list()

        if "SPEMD" in lens_model_list:
            index = lens_model_list.index("SPEMD")
        elif "PEMD" in lens_model_list:
            index = lens_model_list.index("PEMD")
        elif "SPEP" in lens_model_list:
            index = lens_model_list.index("SPEP")
        elif "EPL" in lens_model_list:
            index = lens_model_list.index("EPL")
        else:
            index = None

        return index

    def _get_external_shear_model_index(self):
        """Get the index of the external shear model, if included in the lens model
        list.

        :return: index or `None`
        :rtype: `int`
        """
        lens_model_list = self._config.get_lens_model_list()
        if "SHEAR_GAMMA_PSI" in lens_model_list or "SHEAR" in lens_model_list:
            if "SHEAR_GAMMA_PSI" in lens_model_list:
                index = lens_model_list.index("SHEAR_GAMMA_PSI")
            else:
                index = lens_model_list.index("SHEAR")
        else:
            index = None

        return index

    def _get_shapelet_model_index(self):
        """Get the index of the shapelets model, if included in the source model list.

        :return: index or `None`
        :rtype: `int`
        """
        source_model_list = self._config.get_source_light_model_list()
        if "SHAPELETS" in source_model_list:
            index = source_model_list.index("SHAPELETS")
        else:
            index = None

        return index

    def get_galaxy_quasar_recipe(self):
        """Get the default pre-sampling optimization routine.

        :return: fitting kwargs list
        :rtype: `list`
        """
        fitting_kwargs_list = []

        if self.do_pso:
            pso_range_multipliers = [1.0, 0.1, 0.1]

            pl_model_index = self._get_power_law_model_index()

            for epoch in range(2):
                if epoch == 0 and pl_model_index is not None:
                    fitting_kwargs_list.append(
                        [
                            "update_settings",
                            {"lens_add_fixed": [[pl_model_index, ["gamma"]]]},
                        ]
                    )
                elif pl_model_index is not None:
                    fitting_kwargs_list.append(
                        [
                            "update_settings",
                            {"lens_remove_fixed": [[pl_model_index, ["gamma"]]]},
                        ]
                    )

                for multiplier in pso_range_multipliers:
                    # if multiplier in [10., 1.]:
                    #     fitting_kwargs_list.append([
                    #         'update_settings',
                    #         {'lens_add_fixed': [[index, ['gamma']]]}
                    #
                    #     ])
                    # elif multiplier == .1:
                    #     fitting_kwargs_list.append([
                    #         'update_settings',
                    #         {'lens_remove_fixed': [[index, ['gamma']]]}
                    #     ])

                    fitting_kwargs_list.append(
                        [
                            "PSO",
                            {
                                "sigma_scale": multiplier,
                                "n_particles": self._pso_num_particle,
                                "n_iterations": self._pso_num_iteration,
                                "threadCount": self._thread_count,
                            },
                        ]
                    )

                    if self.reconstruct_psf:
                        fitting_kwargs_list.append(
                            ["psf_iteration", self._config.get_kwargs_psf_iteration()]
                        )
        return fitting_kwargs_list

    def get_sampling_sequence(self):
        """Get the sampling sequence. Currently only MCMC with emcee is supported.

        :return:
        :rtype:
        """
        fitting_kwargs_list = []

        if self.do_sampling:
            supported_samplers = [
                "emcee",
                # "zeus",
                # "dynesty",
                # "dyPolyChord",
                # "MultiNest",
                # "nested_sampling",
                # "Nautilus",
            ]
            if self._config.settings["fitting"]["sampler"] not in supported_samplers:
                raise ValueError(
                    "Sampler '{}' not supported! ".format(
                        self._config.settings["fitting"]["sampler"]
                    )
                    + "Supported ones are: {}".format(supported_samplers)
                )

            sampling_kwargs = self._config.settings["fitting"]["sampler_settings"]
            if self._config.settings["fitting"]["sampler"] in ["emcee"]:
                if "threadCount" not in sampling_kwargs:
                    sampling_kwargs["threadCount"] = self._thread_count

                try:
                    self._config.settings["fitting"]["sampler_settings"]["init_samples"]
                except (NameError, KeyError):
                    pass
                else:
                    if (
                        self._config.settings["fitting"]["sampler_settings"][
                            "init_samples"
                        ]
                        is not None
                    ):
                        sampling_kwargs["init_samples"] = np.array(
                            self._config.settings["fitting"]["sampler_settings"][
                                "init_samples"
                            ]
                        )

                        sampling_kwargs["re_use_samples"] = True

            fitting_kwargs_list.append(
                [self._config.settings["fitting"]["sampler"], sampling_kwargs]
            )

        return fitting_kwargs_list

    def get_galaxy_galaxy_recipe(self, kwargs_data_joint, epochs=2):
        """Get the pre-sampling optimization routine for a galaxy-galaxy lens. PSF
        iteration is not added.

        :param kwargs_data_joint:
        :type kwargs_data_joint:
        :param epochs: number of times to repeat the fitting sequence
        :type epochs: `int`
        :return:
        :rtype:
        """
        fitting_kwargs_list = []

        if self.do_pso:
            arc_masks = []
            masks = self._config.get_masks()
            for i, band_item in enumerate(kwargs_data_joint["multi_band_list"]):
                mask = masks[i] if masks is not None else None
                image = band_item[0]["image_data"]
                arc_masks.append(self.get_arc_mask(image, mask=mask))

            pl_model_index = self._get_power_law_model_index()
            external_shear_model_index = self._get_external_shear_model_index()
            shapelets_index = self._get_shapelet_model_index()

            # temp_constraints = self._config.get_kwargs_constraints()
            for epoch in range(epochs):
                # first fix everything else except for lens light and use arc
                # mask to fit the lens light only. Join the centroids of lens
                # and lens light
                fitting_kwargs_list += [
                    self.fix_params("lens"),
                    self.fix_params("source"),
                    [
                        "update_settings",
                        {
                            "kwargs_likelihood": {
                                "image_likelihood_mask_list": arc_masks
                            }
                        },
                    ],
                    # [
                    #     "update_settings",
                    #     {
                    #         "kwargs_constraints": {
                    #             "joint_lens_with_light": [
                    #                 [0, 0, ["center_x", "center_y"]]
                    #             ]
                    #         }
                    #     },
                    # ],
                    [
                        "PSO",
                        {
                            "sigma_scale": 1.0,
                            "n_particles": self._pso_num_particle,
                            "n_iterations": self._pso_num_iteration,
                            "threadCount": self._thread_count,
                        },
                    ],
                ]

                # unfix the source except for beta, keep lens fixed, fix lens
                # light, use regular mask
                fitting_kwargs_list += [self.unfix_params("source")]

                # fix the shapelets beta parameter
                if shapelets_index is not None:
                    fitting_kwargs_list += [
                        [
                            "update_settings",
                            {"source_add_fixed": [[shapelets_index, ["beta"], [0.1]]]},
                        ]
                    ]

                fitting_kwargs_list += [
                    # self.unfix_params('lens'),
                    self.fix_params("lens_light"),
                    [
                        "update_settings",
                        {"kwargs_likelihood": {"image_likelihood_mask_list": masks}},
                    ],
                ]

                # set lens parameter values to guess values, if provided
                if self.guess_params["lens"] is not None:
                    param_list = []
                    for index, params in self.guess_params["lens"].items():
                        param_list.append(
                            [index, list(params.keys()), list(params.values())]
                        )

                    fitting_kwargs_list += [
                        ["update_settings", {"lens_add_fixed": param_list}]
                    ]

                # optimize for the source only
                fitting_kwargs_list += [
                    # self.fix_params('lens'),
                    [
                        "PSO",
                        {
                            "sigma_scale": 1.0,
                            "n_particles": self._pso_num_particle,
                            "n_iterations": self._pso_num_iteration,
                            "threadCount": self._thread_count,
                        },
                    ],
                ]

                # unfix the central deflector parameters, keep beta fixed
                fitting_kwargs_list += [
                    self.unfix_params("lens"),
                    self.fix_params("lens", external_shear_model_index),
                ]

                # optimize for lens and source together, fix power-law gamma to
                # 2, as all the lens parameters are unfixed
                if pl_model_index is not None:
                    fitting_kwargs_list += [
                        [
                            "update_settings",
                            {"lens_add_fixed": [[pl_model_index, ["gamma"], [2.0]]]},
                        ]
                    ]

                fitting_kwargs_list += [
                    [
                        "PSO",
                        {
                            "sigma_scale": 1.0,
                            "n_particles": self._pso_num_particle,
                            "n_iterations": self._pso_num_iteration,
                            "threadCount": self._thread_count,
                        },
                    ],
                ]

                # unfix the shapelets beta parameter
                if shapelets_index is not None:
                    fitting_kwargs_list += [
                        [
                            "update_settings",
                            {"source_remove_fixed": [[shapelets_index, ["beta"]]]},
                        ]
                    ]

                # finally optimize with all of lens, lens light and source free
                fitting_kwargs_list += [
                    [
                        "PSO",
                        {
                            "sigma_scale": 1.0,
                            "n_particles": self._pso_num_particle,
                            "n_iterations": self._pso_num_iteration,
                            "threadCount": self._thread_count,
                        },
                    ],
                    self.unfix_params("lens_light"),
                    [
                        "PSO",
                        {
                            "sigma_scale": 1.0,
                            "n_particles": self._pso_num_particle,
                            "n_iterations": self._pso_num_iteration,
                            "threadCount": self._thread_count,
                        },
                    ],
                ]

                # finally, relax shear parameters for MCMC later
                # disjoin lens and lens light centroids
                fitting_kwargs_list += [
                    self.unfix_params("lens"),
                    # ["update_settings", {"kwargs_constraints": temp_constraints}],
                ]

            # fitting_kwargs_list += self.get_default_recipe()

        return fitting_kwargs_list

    def get_arc_mask(self, image, clear_center=0.4, mask=None):
        """Create a mask for lensed galaxy arcs from the image of the lens. The lens
        galaxy is required to be close to the center (within a few pixels) of the image.

        :param image: image of the lensing system
        :type image: `ndarray`
        :param clear_center: radius of the central region to **not** mask
        :type clear_center: `float`
        :param mask: a mask to multiply with the arc mask. If the central
            region is masked out in `mask`, then a circle with radius
            `clear_center` will be unmasked.
        :type mask: `ndarray`
        :return: mask for the lensed galaxy arcs
        :rtype: `ndarray`
        """
        # take x- and y- gradient of the image
        x_diff = np.diff(image, axis=1)[1:, :]
        y_diff = np.diff(image, axis=0)[:, 1:]

        w = len(x_diff) - 1
        x, y = np.meshgrid(
            np.linspace(-w / 2, w / 2, int(w + 1)),
            np.linspace(-w / 2, w / 2, int(w + 1)),
        )
        r = np.sqrt(x * x + y * y)

        # compute the radial gradient of the image
        softening = 1e-10
        radial_gradient = -(x_diff * x / (r + softening) + y_diff * y / (r + softening))

        # convert radial_gradient to binary map (+ve to 0 and -ve to 1).
        # where the arc starts when going radially outward, the gradient
        # will be +ve, so this operation marks the inner edge of the arcs
        radial_gradient[np.isnan(radial_gradient)] = 0
        radial_gradient[radial_gradient > 0] = 1
        radial_gradient[radial_gradient <= 0] = 0
        radial_gradient = 1 - radial_gradient

        # unmark any marked pixels from the central region
        radial_gradient[r < int(clear_center / np.max(self._config.pixel_size))] = 0

        # remove connected regions with area less than 5 pixels to remove
        # masked regions created by noise
        structure = np.ones((3, 3))
        filtered_map = deepcopy(radial_gradient)
        id_regions, num_ids = ndimage.label(filtered_map, structure=structure)
        id_sizes = np.array(
            ndimage.sum(radial_gradient, id_regions, range(num_ids + 1))
        )
        area_mask = id_sizes < 5
        filtered_map[area_mask[id_regions]] = 0

        # dilate the binary marked-pixel map
        dilated = np.zeros_like(filtered_map)

        # create structural elements for dilating the image radially outward
        # in each of the four quadrants separately
        a2 = np.tril(np.ones((8, 8)))
        np.fill_diagonal(a2, 0)
        a2 = np.flip(a2, axis=1)  # 1's at lower than anti-diagonal
        a4 = np.flip(a2)  # 1's at upper than anti-diagonal
        a3 = np.rot90(a2)  # 1's at upper than the diagonal
        a1 = np.flip(a3)  # 1's at lower than the diagonal

        dilated[:50, :50] = ndimage.binary_dilation(filtered_map[:50, :50], a4)
        dilated[:50, 50:] = ndimage.binary_dilation(filtered_map[:50, 50:], a3)
        dilated[50:, :50] = ndimage.binary_dilation(filtered_map[50:, :50], a1)
        dilated[50:, 50:] = ndimage.binary_dilation(filtered_map[50:, 50:], a2)

        # increase the size by 1 along both axes to match the image size
        # the mask is the negative of the marked pixel-map
        arc_mask = 1 - np.pad(dilated, ((0, 1), (0, 1)), "minimum")

        # check for bad values
        arc_mask[arc_mask > 0] = 1
        arc_mask[arc_mask <= 0] = 0

        if mask is not None:
            arc_mask *= mask

            w = len(arc_mask) - 1
            x, y = np.meshgrid(
                np.linspace(-w / 2, w / 2, int(w + 1)),
                np.linspace(-w / 2, w / 2, int(w + 1)),
            )
            r = np.sqrt(x * x + y * y)

            arc_mask[r < int(clear_center / np.max(self._config.pixel_size))] = 1

        return arc_mask

    def fix_params(self, model_component, index=None):
        """Fix all the params in `name` that are not fixed by settings.

        :param model_component: name of params type, e.g., 'lens_model'
        :type model_component: `str`
        :param index: profile indices, if `None` all will be fixed
        :type index: `list`
        :return: formatted fit-sequence code to go into `fitting_kwargs_list`
        :rtype: `list`
        """
        if model_component == "lens":
            kwargs_params = self._config.get_lens_model_params()
        # elif model_component == 'point_source':
        #    kwargs_params = self.get_point_source_params()
        elif model_component == "lens_light":
            kwargs_params = self._config.get_lens_light_model_params()
        elif model_component == "source":
            kwargs_params = self._config.get_source_light_model_params()
        else:
            raise ValueError(
                "{} not recognized! Must be lens or "
                "lens_light or source.".format(model_component)
            )

        lower_list = kwargs_params[3]
        fixed_list = kwargs_params[2]

        if index is None:
            index = [i for i, _ in enumerate(lower_list)]

        if not isinstance(index, list):
            index = [index]

        param_list_with_index = []

        for i, (sigma, fixed) in enumerate(zip(lower_list, fixed_list)):
            if i in index:
                param_list = []
                for key, value in sigma.items():
                    if key not in fixed:
                        param_list.append(key)

                param_list_with_index.append([i, param_list])

        key = "{}_add_fixed".format(model_component)

        return ["update_settings", {key: param_list_with_index}]

    def unfix_params(self, model_component, index=None):
        """Unfix all the params in `name` that are not fixed from settings.

        :param model_component: name of params type, e.g., 'lens_model'
        :type model_component: `str`
        :param index: profile indices, if `None` all will be unfixed
        :type index: `list`
        :return: formatted fit-sequence code to go into `fitting_kwargs_list`
        :rtype: `list`
        """
        code = self.fix_params(model_component, index=index)

        old_key = "{}_add_fixed".format(model_component)
        key = "{}_remove_fixed".format(model_component)

        code[1][key] = deepcopy(code[1][old_key])
        del code[1][old_key]

        return code



# -*- coding: utf-8 -*-
"""This module provides classes to post process a model run output."""
__author__ = "ajshajib"

import numpy as np
import matplotlib.pyplot as plt

from lenstronomy.Plots.model_plot import ModelPlot
from lenstronomy.Sampling.parameters import Param
from lenstronomy.Util.class_creator import create_im_sim
from lenstronomy.LensModel.lens_model import LensModel
from lenstronomy.LensModel.Solver.lens_equation_solver import LensEquationSolver
from lenstronomy.Data.coord_transforms import Coordinates
from lenstronomy.LensModel.lens_model_extensions import LensModelExtensions

from dolphin.processor import Processor
from dolphin.processor.config import ModelConfig


class Output(Processor):
    """This class lets."""

    def __init__(self, io_directory):
        """

        :param io_directory: path to the input/output directory. Should not
            end with slash.
        :type io_directory: `str`
        """
        super().__init__(io_directory)

        self._fit_output = None
        self._kwargs_result = None
        self._model_settings = None
        self._samples_mcmc = None
        self._params_mcmc = None

    @property
    def fit_output(self):
        """The output from the `lenstronomy...Fitting_sequence.fit_sequence()`.

        :return:
        :rtype:
        """
        if self._fit_output is None:
            raise ValueError(
                "Model output not specified!"
                "Load an output using the `load_output()`"
                "method!"
            )
        else:
            return self._fit_output

    @property
    def kwargs_result(self):
        """The `kwargs_result` after running a model by calling
        `lenstronomy...Fitting_sequence.fit_sequence()`.

        :return:
        :rtype:
        """
        if self._kwargs_result is None:
            raise ValueError(
                "Model output not specified!"
                "Load an output using the `load_output()`"
                "method!"
            )
        else:
            return self._kwargs_result

    @property
    def model_settings(self):
        """The `kwargs_result` after running a model by calling
        `lenstronomy...Fitting_sequence.fit_sequence()`.

        :return:
        :rtype:
        """
        if self._model_settings is None:
            raise ValueError(
                "Model output not specified!"
                "Load an output using the `load_output()`"
                "method!"
            )
        else:
            return self._model_settings

    @property
    def samples_mcmc(self):
        """The array of MCMC samples from the model run.

        :return:
        :rtype:
        """
        if self._samples_mcmc is None:
            return []
        else:
            return self._samples_mcmc

    @property
    def params_mcmc(self):
        """The non-linear parameters sampled with MCMC.

        :return:
        :rtype:
        """
        if self._params_mcmc is None:
            return []
        else:
            return self._params_mcmc

    @property
    def num_params_mcmc(self):
        """Number of sampled non-linear parameters in MCMC.

        :return:
        :rtype:
        """
        if self._params_mcmc is None:
            return 0
        else:
            return len(self._params_mcmc)

    def swim(self, *args, **kwargs):
        """Override the `swim` method of the `Processor` class to make it not
        callable."""
        raise NotImplementedError

    def load_output(self, lens_name, model_id):
        """Load output from file and save in class variables.

        :param lens_name: lens name
        :type lens_name: `str`
        :param model_id: model identifier provided at run initiation
        :type model_id: `str`
        :return: output dictionary
        :rtype: `dict`
        """
        output = self.file_system.load_output(lens_name, model_id)

        self._model_settings = output["settings"]
        self._kwargs_result = output["kwargs_result"]
        self._fit_output = output["fit_output"]
        self._multi_band_list_out = output["multi_band_list_out"]

        if self.fit_output[-1][0] == "emcee":
            self._samples_mcmc = self.fit_output[-1][1]
            self._params_mcmc = self.fit_output[-1][2]

        return output

    def get_model_plot_instance(
        self,
        lens_name,
        model_id=None,
        kwargs_result=None,
        band_index=0,
        data_cmap="cubehelix",
    ):
        """Get the `ModelPlot` instance from lenstronomy for the lens.

        :param lens_name: name of the lens
        :type lens_name: `str`
        :param model_id: model run identifier
        :type model_id: `str`
        :param kwargs_result: lenstronomy `kwargs_result` dictionary. If
            provided, it will be used to plot the model, otherwise the model
            will be plotted from the saved/loaded outputs for `lens_name` and
            `model_id`.
        :type kwargs_result: `dict`
        :param band_index: index of band to plot for multi-band case
        :type band_index: `int`
        :param data_cmap: colormap for image, reconstruction, and source plots
        :type data_cmap: `str` or `matplotlib.colors.Colormap`
        :return: `ModelPlot` instance, maximum pixel value of the image
        :rtype: `obj`, `float`
        """
        if model_id is None and kwargs_result is None:
            raise ValueError(
                "Either the `model_id` or the `kwargs_result` " "needs to be provided!"
            )

        if kwargs_result is None:
            self.load_output(lens_name, model_id)
            kwargs_result = self.kwargs_result

        multi_band_list_out = self._multi_band_list_out

        config = ModelConfig(
            lens_name=lens_name,
            io_directory=self.io_directory,
            settings=self.model_settings,
        )

        mask = config.get_masks()
        kwargs_model = config.get_kwargs_model()

        v_max = np.log10(multi_band_list_out[band_index][0]["image_data"].max())

        model_plot = ModelPlot(
            multi_band_list_out,
            kwargs_model,
            kwargs_result,
            arrow_size=0.02,
            cmap_string=data_cmap,
            image_likelihood_mask_list=mask,
            multi_band_type="multi-linear",
        )
        return model_plot, v_max

    def plot_model_overview(
        self,
        lens_name,
        model_id=None,
        kwargs_result=None,
        band_index=0,
        data_cmap="cubehelix",
        residual_cmap="RdBu_r",
        convergence_cmap="afmhot",
        magnification_cmap="viridis",
        v_min=None,
        v_max=None,
        source_v_min=None,
        source_v_max=None,
        print_results=False,
        show_source_light=False,
    ):
        """Plot the model, residual, reconstructed source, convergence, and
        magnification profiles. Either `model_id` or `kwargs_result` needs to be
        provided. `kwargs_result` is prioritized for plotting if both are provided.

        :param lens_name: name of the lens
        :type lens_name: `str`
        :param model_id: model run identifier
        :type model_id: `str`
        :param kwargs_result: lenstronomy `kwargs_result` dictionary. If
            provided, it will be used to plot the model, otherwise the model
            will be plotted from the saved/loaded outputs for `lens_name` and
            `model_id`.
        :type kwargs_result: `dict`
        :param band_index: index of band to plot for multi-band case
        :type band_index: `int`
        :param data_cmap: colormap for image, reconstruction, and source plots
        :type data_cmap: `str` or `matplotlib.colors.Colormap`
        :param residual_cmap: colormap for noise residual plot
        :type residual_cmap: `str` or `matplotlib.colors.Colormap`
        :param convergence_cmap: colormap for convergence plot
        :type convergence_cmap: `str` or `matplotlib.colors.Colormap`
        :param magnification_cmap: colormap for magnification plot
        :type magnification_cmap: `str` or `matplotlib.colors.Colormap`
        :param v_min: minimum plotting scale for the model, data, & source plot
        :type v_min: `float` or `int`
        :param v_max: maximum plotting scale for the model, data, & source plot
        :type v_max: `float` or `int`
        :param source_v_min: minimum plotting scale for the source plot
        :type source_v_min: `float` or `int`
        :param source_v_max: maximum plotting scale for the source plot
        :type source_v_max: `float` or `int`
        :param print_results: if true, prints the `kwargs_result` dictionary
        :type print_results: `bool`
        :param show_source_light: if true, replaces convergence plot with
            source light convolved lens decomposition plot and also replaces
            the magnification plot with the source-light subtracted data
            plot
        :type show_source_light: `bool`
        :return: `matplotlib.pyplot.figure` instance with the plots
        :rtype: `matplotlib.pyplot.figure`
        """
        if print_results:
            print_kwargs_result = kwargs_result
            if kwargs_result is None:
                print_kwargs_result = self.load_output(lens_name, model_id)[
                    "kwargs_result"
                ]
            print(print_kwargs_result)

        if v_max is None:
            model_plot, v_max = self.get_model_plot_instance(
                lens_name,
                model_id=model_id,
                kwargs_result=kwargs_result,
                band_index=band_index,
                data_cmap=data_cmap,
            )
        else:
            model_plot = self.get_model_plot_instance(
                lens_name,
                model_id=model_id,
                kwargs_result=kwargs_result,
                band_index=band_index,
                data_cmap=data_cmap,
            )[0]

        fig, axes = plt.subplots(2, 3, figsize=(16, 8))

        model_plot.data_plot(
            ax=axes[0, 0], band_index=band_index, v_max=v_max, v_min=v_min
        )
        model_plot.model_plot(
            ax=axes[0, 1], band_index=band_index, v_max=v_max, v_min=v_min
        )
        model_plot.normalized_residual_plot(
            ax=axes[0, 2], band_index=band_index, cmap=residual_cmap, v_max=3, v_min=-3
        )
        model_plot.source_plot(
            ax=axes[1, 0],
            deltaPix_source=0.02,
            numPix=100,
            band_index=band_index,
            v_max=source_v_max,
            v_min=source_v_min,
            scale_size=0.4,
        )
        if not show_source_light:
            model_plot.convergence_plot(
                ax=axes[1, 1], band_index=band_index, cmap=convergence_cmap
            )
            model_plot.magnification_plot(
                ax=axes[1, 2], band_index=band_index, cmap=magnification_cmap
            )
        else:
            model_plot.subtract_from_data_plot(
                ax=axes[1, 1],
                band_index=band_index,
                lens_light_add=True,
                v_max=v_max,
                v_min=v_min,
            )
            model_plot.decomposition_plot(
                ax=axes[1, 2],
                text="Source light convolved",
                source_add=True,
                band_index=band_index,
                v_max=v_max,
                v_min=v_min,
            )
        fig.tight_layout()
        fig.subplots_adjust(
            left=None, bottom=None, right=None, top=None, wspace=0.0, hspace=0.05
        )

        return fig

    def plot_model_decomposition(
        self,
        lens_name,
        model_id=None,
        kwargs_result=None,
        band_index=0,
        data_cmap="cubehelix",
        v_min=None,
        v_max=None,
    ):
        """Plot lens light and source light model decomposition, both with convolved and
        unconvolved light. Either `model_id` or `kwargs_result` needs to be provided.
        `kwargs_result` is prioritized for plotting if both are provided.

        :param lens_name: name of the lens
        :type lens_name: `str`
        :param model_id: model run identifier
        :type model_id: `str`
        :param kwargs_result: lenstronomy `kwargs_result` dictionary. If
            provided, it will be used to plot the model, otherwise the model
            will be plotted from the saved/loaded outputs for `lens_name` and
            `model_id`.
        :type kwargs_result: `dict`
        :param band_index: index of band to plot for multi-band case
        :type band_index: `int`
        :param data_cmap: colormap for image, reconstruction, and source plots
        :type data_cmap: `str` or `matplotlib.colors.Colormap`
        :param v_min: minimum plotting scale for the model, data, & source plot
        :type v_min: `float` or `int`
        :param v_max: maximum plotting scale for the model, data, & source plot
        :type v_max: `float` or `int`
        :return: `matplotlib.pyplot.figure` instance with the plots
        :rtype: `matplotlib.pyplot.figure`
        """

        if v_max is None:
            model_plot, v_max = self.get_model_plot_instance(
                lens_name,
                model_id=model_id,
                kwargs_result=kwargs_result,
                band_index=band_index,
                data_cmap=data_cmap,
            )
        else:
            model_plot = self.get_model_plot_instance(
                lens_name,
                model_id=model_id,
                kwargs_result=kwargs_result,
                band_index=band_index,
                data_cmap=data_cmap,
            )[0]

        fig, axes = plt.subplots(2, 3, figsize=(16, 8))
        model_plot.decomposition_plot(
            ax=axes[0, 0],
            text="Lens light",
            lens_light_add=True,
            unconvolved=True,
            band_index=band_index,
            v_max=v_max,
            v_min=v_min,
        )
        model_plot.decomposition_plot(
            ax=axes[1, 0],
            text="Lens light convolved",
            lens_light_add=True,
            band_index=band_index,
            v_max=v_max,
            v_min=v_min,
        )
        model_plot.decomposition_plot(
            ax=axes[0, 1],
            text="Source light",
            source_add=True,
            unconvolved=True,
            band_index=band_index,
            v_max=v_max,
            v_min=v_min,
        )
        model_plot.decomposition_plot(
            ax=axes[1, 1],
            text="Source light convolved",
            source_add=True,
            band_index=band_index,
            v_max=v_max,
            v_min=v_min,
        )
        model_plot.decomposition_plot(
            ax=axes[0, 2],
            text="All components",
            source_add=True,
            lens_light_add=True,
            unconvolved=True,
            band_index=band_index,
            v_max=v_max,
            v_min=v_min,
        )
        model_plot.decomposition_plot(
            ax=axes[1, 2],
            text="All components convolved",
            source_add=True,
            lens_light_add=True,
            point_source_add=True,
            band_index=band_index,
            v_max=v_max,
            v_min=v_min,
        )
        fig.tight_layout()
        fig.subplots_adjust(
            left=None, bottom=None, right=None, top=None, wspace=0.0, hspace=0.05
        )

        return fig

    def get_reshaped_emcee_chain(
        self, lens_name, model_id, walker_ratio, burn_in=0, verbose=True
    ):
        """

        :param lens_name:
        :type lens_name:
        :param model_id:
        :type model_id:
        :param walker_ratio:
        :type walker_ratio:
        :param burn_in:
        :type burn_in:
        :param verbose:
        :type verbose:
        :return:
        :rtype:
        """
        self.load_output(lens_name, model_id)

        num_params = self.num_params_mcmc  # self.samples_mcmc.shape[1]
        num_walkers = walker_ratio * num_params
        num_step = int(len(self.samples_mcmc) / num_walkers)

        chain = np.empty((num_walkers, num_step, num_params))

        for i in np.arange(num_params):
            samples = self.samples_mcmc[:, i].T
            chain[:, :, i] = samples.reshape((num_step, num_walkers)).T
        if burn_in != 0:
            chain = chain[:, burn_in:, :]

        return chain

    def plot_mcmc_trace(
        self,
        lens_name,
        model_id,
        walker_ratio,
        burn_in=-100,
        verbose=True,
        fig_width=16,
        parameters_to_plot=[],
    ):
        """Plot the trace of MCMC walkers.

        :param lens_name: name of the lens
        :type lens_name: `str`
        :param model_id: model run identifier
        :type model_id: `str`
        :param walker_ratio: number of walkers per parameter in MCMC
        :type walker_ratio: `int`
        :param burn_in: number of burn-in steps to compute the medians after
            convergence of the MCMC chain
        :type: `int`
        :param verbose: if `True`, median values after burn-in will be printed
        :type verbose: `bool`
        :param fig_width: width of the figure
        :type fig_width: `float`
        :param parameters_to_plot: if not empty, list of parameters to plot
        :type fig_width: `list`
        :return: `matplotlib.pyplot.figure` instance with the plots
        :rtype: `matplotlib.pyplot.figure`
        """
        chain = self.get_reshaped_emcee_chain(
            lens_name, model_id, walker_ratio, burn_in=burn_in, verbose=verbose
        )

        num_params = self.num_params_mcmc
        num_step = chain.shape[1]

        if len(parameters_to_plot) == 0:
            parameter_list = np.arange(num_params)
        else:
            parameter_list = []
            for i in parameters_to_plot:
                if i in self.params_mcmc:
                    parameter_list.append(self.params_mcmc.index(i))
                else:
                    raise ValueError(
                        f"Parameter '{i}' not found. Available parameters: {self.params_mcmc}"
                    )

        mean_pos = np.zeros((num_params, num_step))
        median_pos = np.zeros((num_params, num_step))
        std_pos = np.zeros((num_params, num_step))
        q16_pos = np.zeros((num_params, num_step))
        q84_pos = np.zeros((num_params, num_step))

        # chain = np.empty((nwalker, nstep, ndim), dtype = np.double)
        for i in parameter_list:
            for j in np.arange(num_step):
                mean_pos[i][j] = np.mean(chain[:, j, i])
                median_pos[i][j] = np.median(chain[:, j, i])
                std_pos[i][j] = np.std(chain[:, j, i])
                q16_pos[i][j] = np.percentile(chain[:, j, i], 16.0)
                q84_pos[i][j] = np.percentile(chain[:, j, i], 84.0)

        fig, ax = plt.subplots(
            len(parameter_list),
            sharex="all",
            figsize=(fig_width, int(fig_width / 8) * len(parameter_list)),
        )
        last = num_step
        medians = []

        for n, i in enumerate(parameter_list):
            if verbose:
                print(
                    self.params_mcmc[i],
                    "{:.4f} ± {:.4f}".format(
                        median_pos[i][last - 1],
                        (q84_pos[i][last - 1] - q16_pos[i][last - 1]) / 2,
                    ),
                )
            if len(parameter_list) != 1:
                # ax[i].plot(mean_pos[i][:3000], c='b')
                ax[n].plot(median_pos[i][:last], c="g")
                # ax[i].axhline(np.mean(mean_pos[i][burnin:2900]), c='b')
                ax[n].axhline(np.median(median_pos[i][burn_in:last]), c="r", lw=1)
                ax[n].fill_between(
                    np.arange(last), q84_pos[i][:last], q16_pos[i][:last], alpha=0.4
                )
                # ax[i].fill_between(np.arange(last), mean_pos[i][:last] \
                # +std_pos[i][:last], mean_pos[i][:last]-std_pos[i][:last],
                # alpha=0.4)
                ax[n].set_ylabel(self.params_mcmc[i], fontsize=8)
                ax[n].set_xlim(0, last)
            else:
                ax.plot(median_pos[i][:last], c="g")
                ax.axhline(np.median(median_pos[i][burn_in:last]), c="r", lw=1)
                ax.fill_between(
                    np.arange(last), q84_pos[i][:last], q16_pos[i][:last], alpha=0.4
                )
                ax.set_ylabel(self.params_mcmc[i], fontsize=8)
                ax.set_xlim(0, last)

            medians.append(np.median(median_pos[i][burn_in:last]))

        return fig

    def get_param_class(self, lens_name, model_id):
        """Get `Param` instance for the lens model.

        :param lens_name: name of the lens
        :type lens_name: `str`
        :param model_id: model run identifier
        :type model_id: `str`
        :return: `Param` instance
        :rtype: `obj`
        """
        self.load_output(lens_name, model_id=model_id)

        config = ModelConfig(lens_name=lens_name, settings=self._model_settings)
        kwargs_params = config.get_kwargs_params()
        kwargs_model = config.get_kwargs_model()
        kwargs_constraints = config.get_kwargs_constraints()

        param = Param(
            kwargs_model,
            kwargs_params["lens_model"][2],
            kwargs_params["source_model"][2],
            kwargs_params["lens_light_model"][2],
            kwargs_params["point_source_model"][2],
            # kwargs_params['special'][2],
            # kwargs_params['extinction_model'][2],
            # kwargs_lens_init=kwargs_result['kwargs_lens'],
            kwargs_lens_init=kwargs_params["lens_model"][0],
            **kwargs_constraints,
        )
        return param

    def get_kwargs_from_args(
        self, lens_name, model_id, args, band_index=0, linear_solve=False, param=None
    ):
        """

        :param lens_name:
        :type lens_name:
        :param model_id:
        :type model_id:
        :return:
        :rtype:
        """
        if param is None:
            param = self.get_param_class(lens_name, model_id)

        kwargs = param.args2kwargs(args)

        if linear_solve:
            im_sim = self.get_im_sim(lens_name, band_index)

            im_sim.image_linear_solve(
                kwargs_lens=kwargs["kwargs_lens"],
                kwargs_source=kwargs["kwargs_source"],
                kwargs_lens_light=kwargs["kwargs_lens_light"],
                kwargs_ps=kwargs["kwargs_ps"],
                kwargs_extinction=kwargs["kwargs_extinction"],
                kwargs_special=kwargs["kwargs_special"],
                inv_bool=True,
            )

        return kwargs

    def get_im_sim(self, lens_name, band_index):
        """Get lestronomy's `ImSim` instance for the lens.

        :param lens_name: name of the lens
        :type lens_name: `str`
        :param band_index: index of band to plot for multi-band case
        :type band_index: `int`
        :return: `ImSim` instance
        :rtype: `obj`
        """
        config = ModelConfig(lens_name=lens_name, settings=self._model_settings)

        # kwargs_numerics = config.get_kwargs_numerics()
        kwargs_model = config.get_kwargs_model()

        multi_band_list_out = self._multi_band_list_out

        # kwargs_data = multi_band_list_out[band_index][0]
        # kwargs_psf = multi_band_list_out[band_index][1]

        im_sim = create_im_sim(
            multi_band_list_out,
            "single-band",
            kwargs_model,
            bands_compute=None,
            image_likelihood_mask_list=(config.get_masks()),
            band_index=band_index,
        )

        return im_sim

    def get_magnification_extended_source(
        self,
        lens_name,
        model_id=None,
        kwargs_result=None,
        band_index=0,
        plot=True,
        v_max=None,
        v_min=None,
    ):
        """Get the total magnification of the lens system for an extended source.

        :param lens_name: name of the lens
        :type lens_name: `str`
        :param model_id: model run identifier
        :type model_id: `str`
        :param band_index: index of band to plot for multi-band case
        :type band_index: `int`
        :param plot: if `True`, plots the lensed and unlensed source
        :type plot: `bool`
        :param kwargs_result: lenstronomy `kwargs_result` dictionary. If
            provided, it will be used to make the calculation, otherwise the model
            will be calculated from the saved/loaded outputs for `lens_name` and
            `model_id`.
        :type kwargs_result: `dict`
        :raises ValueError: if neither `model_id` nor `kwargs_result` is provided
        :return: average magnification
        :rtype: `float`
        """
        if model_id is None and kwargs_result is None:
            raise ValueError(
                "Either the `model_id` or the `kwargs_result` " "needs to be provided!"
            )
        if kwargs_result is None:
            self.load_output(lens_name, model_id)
            kwargs_result = self.kwargs_result

        im_sim = self.get_im_sim(lens_name, band_index)

        model_plot = self.get_model_plot_instance(
            lens_name,
            model_id=model_id,
            kwargs_result=kwargs_result,
            band_index=band_index,
        )[0]

        source_lensed = im_sim.image(
            kwargs_result["kwargs_lens"],
            kwargs_result["kwargs_source"],
            kwargs_result["kwargs_lens_light"],
            kwargs_result["kwargs_ps"],
            lens_light_add=False,
            point_source_add=False,
            unconvolved=False,
        )

        config = ModelConfig(lens_name=lens_name, settings=self._model_settings)
        image_pixel_size = config.pixel_size[band_index]

        delta_pix = image_pixel_size / 2
        num_pix = source_lensed.shape[0]
        source_unlensed = model_plot.source(
            deltaPix=delta_pix,
            numPix=num_pix,
            band_index=band_index,
            center=[
                kwargs_result["kwargs_source"][0]["center_x"],
                kwargs_result["kwargs_source"][0]["center_y"],
            ],
        )[0]

        magnification = np.sum(source_lensed) / np.sum(source_unlensed)

        if plot:
            fig = plt.figure()
            ax = fig.add_subplot(1, 2, 1)
            ax.set_title("Lensed source (convolved)")
            ax.imshow(
                np.log10(source_lensed),
                cmap="cubehelix",
                vmax=v_max,
                vmin=v_min,
                interpolation="none",
                origin="lower",
            )
            ax.axis("off")

            ax = fig.add_subplot(1, 2, 2)
            ax.set_title("Unlensed source")
            ax.imshow(
                np.log10(source_unlensed),
                cmap="cubehelix",
                vmax=v_max,
                vmin=v_min,
                interpolation="none",
                origin="lower",
            )
            ax.axis("off")

        return magnification

    def get_critical_curve(
        self,
        lens_name,
        model_id=None,
        kwargs_result=None,
        band_index=0,
    ):
        """Get the critical curve of the lens system.

        :param lens_name: name of the lens
        :type lens_name: `str`
        :param model_id: model run identifier
        :type model_id: `str`
        :param kwargs_result: lenstronomy `kwargs_result` dictionary. If
            provided, it will be used to make the calculation, otherwise the model
            will be calculated from the saved/loaded outputs for `lens_name` and
            `model_id`.
        :type kwargs_result: `dict`
        :raises ValueError: if neither `model_id` nor `kwargs_result` is provided
        :raises ValueError: if `band_index` is out of bounds
        :return: critical curve coordinates
        :rtype: `tuple`
        """
        if model_id is None and kwargs_result is None:
            raise ValueError(
                "Either the `model_id` or the `kwargs_result` " "needs to be provided!"
            )
        if kwargs_result is None:
            self.load_output(lens_name, model_id)
            kwargs_result = self.kwargs_result

        # solve lens equation
        config = ModelConfig(lens_name=lens_name, settings=self._model_settings)
        kwargs_model = config.get_kwargs_model()

        lens_model_class = LensModel(lens_model_list=kwargs_model["lens_model_list"])

        lens_model_extensions = LensModelExtensions(lens_model_class)

        image_pixel_size = config.pixel_size[band_index]
        multi_band_list_out = self._multi_band_list_out
        image_data = multi_band_list_out[band_index][0]["image_data"]
        image_size = image_data.shape[0] * image_pixel_size

        (
            ra_crit_list,
            dec_crit_list,
        ) = lens_model_extensions.critical_curve_tiling(
            kwargs_result["kwargs_lens"],
            compute_window=image_size,
            start_scale=image_pixel_size / 2.0,
            max_order=3,
            center_x=0,
            center_y=0,
        )

        return ra_crit_list, dec_crit_list

    def get_magnification_point_source(
        self,
        lens_name,
        model_id=None,
        kwargs_result=None,
        band_index=0,
        plot=True,
        v_max=None,
        v_min=None,
    ):
        """Get the magnifications for each image for a point source, either isolated or
        that situated at the center of the extended host.

        :param lens_name: name of the lens
        :type lens_name: `str`
        :param model_id: model run identifier
        :type model_id: `str`
        :param kwargs_result: lenstronomy `kwargs_result` dictionary. If
            provided, it will be used to make the calculation, otherwise the model
            will be calculated from the saved/loaded outputs for `lens_name` and
            `model_id`.
        :type kwargs_result: `dict`
        :param band_index: index of band to plot for multi-band case
        :type band_index: `int`
        :param plot: if `True`, plots the lensed and unlensed source
        :type plot: `bool`
        :param v_max: maximum plotting scale for the model, data, & source plot
        :type v_max: `float` or `int`
        :param v_min: minimum plotting scale for the model, data, & source plot
        :type v_min: `float` or `int`
        :raises ValueError: if neither `model_id` nor `kwargs_result` is provided
        :raises ValueError: if `band_index` is out of bounds
        :return: individual magnifications of the images
        :rtype: `list`
        """
        if model_id is None and kwargs_result is None:
            raise ValueError(
                "Either the `model_id` or the `kwargs_result` " "needs to be provided!"
            )
        if kwargs_result is None:
            self.load_output(lens_name, model_id)
            kwargs_result = self.kwargs_result

        # solve lens equation
        config = ModelConfig(lens_name=lens_name, settings=self._model_settings)
        kwargs_model = config.get_kwargs_model()

        lens_model_class = LensModel(lens_model_list=kwargs_model["lens_model_list"])
        lensEquationSolver = LensEquationSolver(lens_model_class)
        x_source = kwargs_result["kwargs_source"][0]["center_x"]
        y_source = kwargs_result["kwargs_source"][0]["center_y"]

        kwargs_lens = kwargs_result["kwargs_lens"]

        image_pixel_size = config.pixel_size[band_index]

        multi_band_list_out = self._multi_band_list_out
        image_data = multi_band_list_out[band_index][0]["image_data"]
        image_size = image_data.shape[0] * image_pixel_size

        x_image, y_image = lensEquationSolver.findBrightImage(
            x_source,
            y_source,
            kwargs_lens,
            numImages=4,
            min_distance=image_pixel_size,
            search_window=image_size,
        )

        magnifications = []
        for x, y in zip(x_image, y_image):
            magnifications.append(lens_model_class.magnification(x, y, kwargs_lens))

        if plot:
            ra_at_xy_0 = multi_band_list_out[band_index][0]["ra_at_xy_0"]
            dec_at_xy_0 = multi_band_list_out[band_index][0]["dec_at_xy_0"]
            transform_pix2angle = multi_band_list_out[band_index][0][
                "transform_pix2angle"
            ]

            coordinate_system = Coordinates(
                transform_pix2angle, ra_at_xy_0, dec_at_xy_0
            )

            ra_crit_list, dec_crit_list = self.get_critical_curve(
                lens_name, model_id=model_id, kwargs_result=kwargs_result
            )
            x_crit, y_crit = coordinate_system.map_coord2pix(
                ra_crit_list, dec_crit_list
            )

            fig = plt.figure()
            ax = fig.add_subplot(1, 1, 1)

            temp_image_data = np.copy(image_data)
            temp_image_data[temp_image_data <= 0] = 1e-6
            ax.imshow(
                np.log10(temp_image_data),
                cmap="cubehelix",
                vmax=v_max,
                vmin=v_min,
                origin="lower",
                interpolation="none",
            )

            ax.scatter(
                x_crit,
                y_crit,
                marker="o",
                color="#ff7f00",
                alpha=0.5,
                s=0.5,
            )

            x_pix, y_pix = coordinate_system.map_coord2pix(x_image, y_image)
            markersize = 50
            for i, ma in enumerate(magnifications):
                ax.scatter(
                    x_pix[i],
                    y_pix[i],
                    marker="d",
                    color="#ffff33",
                    edgecolors="black",
                    alpha=0.6,
                    s=(1 + np.log10(np.abs(ma))) * markersize,
                )
                ax.text(
                    x_pix[i] + 3,
                    y_pix[i],
                    str(round(magnifications[i], 1)),
                    color="#ffff33",
                    ha="left",
                    va="center",
                )

            # for legend
            ax.scatter(
                x_pix[0] * 1000,
                y_pix[0] * 1000,
                marker="d",
                color="#ffff33",
                edgecolors="black",
                alpha=0.6,
                s=markersize,
                label="image",
            )

            source_x_pix, source_y_pix = coordinate_system.map_coord2pix(
                x_source, y_source
            )
            ax.scatter(
                source_x_pix,
                source_y_pix,
                marker="o",
                color="#e41a1c",
                edgecolors="black",
                alpha=0.6,
                s=markersize,
                label="source",
            )

            ax.axis("off")
            ax.set_xlim(0, image_data.shape[0])
            ax.set_ylim(0, image_data.shape[1])

            ax.set_title(lens_name)

            ax.legend(loc="upper right", framealpha=0.9)

            plt.show()

        return magnifications



# -*- coding: utf-8 -*-
"""This module loads settings from a configuration file."""
__author__ = "ajshajib"

import yaml
import numpy as np
from copy import deepcopy

from lenstronomy.Util.param_util import ellipticity2phi_q
import lenstronomy.Util.util as util
import lenstronomy.Util.mask_util as mask_util
import os

from .data import ImageData
from .files import FileSystem


class Config(object):
    """This class contains the methods to load an read YAML configuration files.

    This is a parent class for other classes that needs to load a configuration file. If
    the file type of the configuration files changes, then only this class needs to be
    modified.
    """

    def __init__(self):
        pass

    @classmethod
    def load_config_from_yaml(cls, file):
        """Load configuration from `file`.

        :return:
        :rtype:
        """
        with open(file, "r") as f:
            settings = yaml.load(f, yaml.FullLoader)

        return settings


class ModelConfig(Config):
    """This class contains the methods to load and interact with modeling settings for a
    particular system."""

    def __init__(self, lens_name, file_system=None, io_directory=None, settings=None):
        """Initiate a Model Config object. If the file path is given, `settings` will be
        loaded from it. Otherwise, the `settings` can be loaded/reloaded later with the
        `load_settings_from_file` method.

        :param lens_name: name of the lens system
        :type lens_name: `str`
        :param file_system: a FileSystem object
        :type file_system: `FileSystem`
        :param io_directory: path to the input-output directory
        :type io_directory: `str`
        :param settings: a dictionary containing settings. If both `file`
            and `settings` are provided, `file` will be prioritized.
        :type settings: `dict`
        """
        super().__init__()

        if file_system is not None:
            self._file_system = file_system
        elif io_directory is not None:
            self._file_system = FileSystem(io_directory)

        self._lens_name = lens_name

        if settings is not None:
            self.settings = settings
        else:
            self._config_file_path = self._file_system.get_config_file_path(lens_name)

            self._settings_dir = os.path.dirname(self._config_file_path)
            self.settings = self.load_config_from_yaml(self._config_file_path)

        assert self.settings["lens_name"] == self._lens_name

    @property
    def lens_name(self):
        """The name of the lens system.

        :return:
        :rtype:
        """
        return self._lens_name

    @property
    def pixel_size(self):
        """The pixel size.

        :return:
        :rtype:
        """
        # if isinstance(self.settings["pixel_size"], float):
        #     return [self.settings["pixel_size"]] * self.band_number
        # else:
        #     return self.settings["pixel_size"]
        if "pixel_size" not in self.settings:
            pixel_size = []

            for band in self.settings["band"]:
                image_data = self.get_image_data(band)
                pixel_size.append(image_data.get_image_pixel_scale())

            self.settings["pixel_size"] = pixel_size

        return self.settings["pixel_size"]

    def get_image_data(self, band):
        """Get image data.

        :param band: name of the band
        :type band: `str`
        :return: image data
        :rtype: `ImageData`
        """
        return ImageData(self._file_system.get_image_file_path(self.lens_name, band))

    @property
    def deflector_center_ra(self):
        """The RA offset for the deflector's center from the zero-point in the
        coordinate system of the data. Default is 0.

        :return:
        :rtype:
        """
        if (
            "lens_option" in self.settings
            and "centroid_init" in self.settings["lens_option"]
        ):
            return float(self.settings["lens_option"]["centroid_init"][0])
        else:
            return 0.0

    @property
    def deflector_center_dec(self):
        """The dec offset for the deflector's center from the zero-point in the
        coordinate system of the data. Default is 0.

        :return:
        :rtype:
        """
        if (
            "lens_option" in self.settings
            and "centroid_init" in self.settings["lens_option"]
        ):
            return float(self.settings["lens_option"]["centroid_init"][1])
        else:
            return 0.0

    @property
    def deflector_centroid_bound(self):
        """Half of the box width to constrain the deflector's centroid. Default is 0.5
        arcsec.

        :return:
        :rtype:
        """
        if "lens_option" in self.settings:
            if "centroid_bound" in self.settings["lens_option"]:
                bound = self.settings["lens_option"]["centroid_bound"]
                if bound is not None:
                    return bound

        return 0.5

    @property
    def number_of_bands(self):
        """The number of bands.

        :return:
        :rtype:
        """
        return len(self.settings["band"])

    def get_kwargs_model(self):
        """Create `kwargs_model`.

        :return:
        :rtype:
        """
        lens_model_list = deepcopy(self.get_lens_model_list())
        if "SIE" in lens_model_list:
            index = lens_model_list.index("SIE")
            lens_model_list[index] = "EPL"

        kwargs_model = {
            "lens_model_list": lens_model_list,
            "source_light_model_list": self.get_source_light_model_list(),
            "lens_light_model_list": self.get_lens_light_model_list(),
            "point_source_model_list": self.get_point_source_model_list(),
            "index_lens_light_model_list": self.get_index_lens_light_model_list(),
            "index_source_light_model_list": self.get_index_source_light_model_list(),
        }

        if (
            "kwargs_model" in self.settings
            and self.settings["kwargs_model"] is not None
        ):
            for key, value in self.settings["kwargs_model"].items():
                kwargs_model[key] = value

        return kwargs_model

    def get_kwargs_constraints(self):
        """Create `kwargs_constraints`.

        :return:
        :rtype:
        """
        joint_source_with_source, num_source_profiles = (
            self.get_joint_source_with_source()
        )

        joint_lens_light_with_lens_light = self.get_joint_lens_light_with_lens_light()

        joint_source_with_point_source = self.get_joint_source_with_point_source(
            num_source_profiles
        )

        joint_lens_with_light = self.get_joint_lens_with_light()

        kwargs_constraints = {
            "joint_source_with_source": joint_source_with_source,
            "joint_lens_light_with_lens_light": joint_lens_light_with_lens_light,
            "joint_source_with_point_source": joint_source_with_point_source,
            "joint_lens_with_light": joint_lens_with_light,
            "joint_lens_with_lens": [],
        }

        if len(self.get_point_source_model_list()) > 0:
            num_image = len(self.settings["point_source_option"]["ra_init"])
            kwargs_constraints["num_point_source_list"] = [num_image]
            kwargs_constraints["solver_type"] = (
                "PROFILE_SHEAR" if num_image > 2 else "CENTER"
            )

        if (
            "kwargs_constraints" in self.settings
            and self.settings["kwargs_constraints"] is not None
        ):
            for key, value in self.settings["kwargs_constraints"].items():
                kwargs_constraints[key] = value

        return kwargs_constraints

    def get_joint_lens_with_light(self):
        """Create `joint_lens_with_light`."""
        _, lens_light_satellite_flags = self.get_lens_light_model_list_with_flags()
        _, lens_satellite_flags = self.get_lens_model_list_with_flags()

        # join centers of main deflector's mass and light
        joint_lens_with_light = [[0, 0, ["center_x", "center_y"]]]

        # joint satellite mass and light
        if (np.array(lens_light_satellite_flags) > -1).any():
            for i, flag in enumerate(lens_light_satellite_flags):
                if flag > -1:
                    lens_sat_index = lens_satellite_flags.index(flag)
                    joint_lens_with_light.append(
                        [i, lens_sat_index, ["center_x", "center_y"]]
                    )

        return joint_lens_with_light

    def get_joint_source_with_point_source(self, num_source_profiles):
        """Create `joint_source_with_point_source`."""
        joint_source_with_point_source = []
        if len(self.get_point_source_model_list()) > 0 and num_source_profiles > 0:
            for n in range(num_source_profiles):
                joint_source_with_point_source.append([0, n])
        return joint_source_with_point_source

    def get_joint_lens_light_with_lens_light(self):
        """Create `joint_lens_light_with_lens_light`."""
        joint_lens_light_with_lens_light = []
        lens_light_model_list = self.get_lens_light_model_list()

        num_lens_light_profile_central = len(self.settings["model"]["lens_light"])

        if num_lens_light_profile_central > 1:
            for n in range(1, num_lens_light_profile_central * self.number_of_bands):
                joint_lens_light_with_lens_light.append(
                    [0, n, ["center_x", "center_y"]]
                )

        # Join Sersic ellipticities in multiband fitting
        if self.number_of_bands > 1:
            for i in range(num_lens_light_profile_central):
                model = lens_light_model_list[i]
                if "SERSIC" in model:
                    join_list = ["n_sersic"]
                    if "ELLIPSE" in model:
                        join_list += ["e1", "e2"]
                    joint_lens_light_with_lens_light.append(
                        [
                            i,
                            i + num_lens_light_profile_central,
                            join_list,
                        ]
                    )

        if self.number_of_bands > 1 and self.num_satellites > 0:
            for i in range(self.num_satellites):
                model = lens_light_model_list[i + num_lens_light_profile_central]

                join_list = ["center_x", "center_y", "n_sersic"]
                if "ELLIPSE" in model:
                    join_list += ["e1", "e2"]
                joint_lens_light_with_lens_light.append(
                    [
                        i + num_lens_light_profile_central * self.number_of_bands,
                        i
                        + num_lens_light_profile_central * self.number_of_bands
                        + self.num_satellites,
                        join_list,
                    ]
                )

        return joint_lens_light_with_lens_light

    def get_joint_source_with_source(self):
        """Create `joint_source_with_source`."""
        joint_source_with_source = []
        num_source_profiles = len(self.get_source_light_model_list())

        if num_source_profiles > 1:
            for n in range(1, num_source_profiles):
                joint_source_with_source.append([0, n, ["center_x", "center_y"]])

        # Join Sersic ellipticities in multiband fitting
        num_bands = self.number_of_bands
        if num_bands > 1:
            num_source_profile_single_band = int(num_source_profiles / num_bands)

            for i in range(num_source_profile_single_band):
                model = self.get_source_light_model_list()[i]
                if "SERSIC" in model:
                    join_list = ["n_sersic"]
                    if "ELLIPSE" in model:
                        join_list += ["e1", "e2"]
                    joint_source_with_source.append(
                        [
                            i,
                            i + num_source_profile_single_band,
                            join_list,
                        ]
                    )

        return joint_source_with_source, num_source_profiles

    def get_kwargs_likelihood(self):
        """Create `kwargs_likelihood`.

        :return:
        :rtype:
        """
        kwargs_likelihood = {
            "force_no_add_image": False,
            "source_marg": False,
            # 'point_source_likelihood': True,
            # 'position_uncertainty': 0.00004,
            # 'check_solver': False,
            # 'solver_tolerance': 0.001,
            "check_positive_flux": True,
            "check_bounds": True,
            "bands_compute": [True] * self.number_of_bands,
            "image_likelihood_mask_list": self.get_masks(),
            "prior_lens": [],
            "prior_lens_light": [],
            "prior_ps": [],
            "prior_source": [],
        }

        if (
            "lens_option" in self.settings
            and "gaussian_prior" in self.settings["lens_option"]
        ):
            for index, param_dict in self.settings["lens_option"][
                "gaussian_prior"
            ].items():
                for i in param_dict:
                    prior_param = [index]
                    prior_param.extend(i)
                    kwargs_likelihood["prior_lens"].append(prior_param)

        if (
            "lens_light_option" in self.settings
            and "gaussian_prior" in self.settings["lens_light_option"]
        ):
            for index, param_dict in self.settings["lens_light_option"][
                "gaussian_prior"
            ].items():
                for i in param_dict:
                    prior_param = [index]
                    prior_param.extend(i)
                    kwargs_likelihood["prior_lens_light"].append(prior_param)

        if (
            "source_light_option" in self.settings
            and "gaussian_prior" in self.settings["source_light_option"]
        ):
            for index, param_dict in self.settings["source_light_option"][
                "gaussian_prior"
            ].items():
                for i in param_dict:
                    prior_param = [index]
                    prior_param.extend(i)
                    kwargs_likelihood["prior_source"].append(prior_param)

        if (
            "point_source_option" in self.settings
            and "gaussian_prior" in self.settings["point_source_option"]
        ):
            for index, param_dict in self.settings["point_source_option"][
                "gaussian_prior"
            ].items():
                for i in param_dict:
                    prior_param = [index]
                    prior_param.extend(i)
                    kwargs_likelihood["prior_ps"].append(prior_param)

        use_custom_logL_addition = False

        if "lens_option" in self.settings:
            if any(
                key in self.settings["lens_option"]
                for key in ["limit_mass_pa_from_light", "limit_mass_q_from_light"]
            ):
                use_custom_logL_addition = True

        if "source_light_option" in self.settings:
            if (
                "shapelet_scale_logarithmic_prior"
                in self.settings["source_light_option"]
            ):
                use_custom_logL_addition = True

        if use_custom_logL_addition:
            kwargs_likelihood["custom_logL_addition"] = self.custom_logL_addition

        return kwargs_likelihood

    def custom_logL_addition(
        self,
        kwargs_lens=None,
        kwargs_source=None,
        kwargs_lens_light=None,
        kwargs_ps=None,
        kwargs_special=None,
        kwargs_extinction=None,
        kwargs_tracer_source=None,
    ):
        """Provide additional likelihood terms to be sent to `lenstronomy`.

        :param kwargs_lens: dictionary containing lens model keyword arguments
        :type kwargs_lens: `dict`
        :param kwargs_source: dictionary containing source model keyword arguments
        :type kwargs_source: `dict`
        :param kwargs_lens_light: dictionary containing lens light model keyword arguments
        :type kwargs_lens_light: `dict`
        :param kwargs_ps: dictionary containing point source model keyword arguments
        :type kwargs_ps: `dict`
        :param kwargs_special: dictionary containing special model keyword arguments
        :type kwargs_special: `dict`
        :param kwargs_extinction: dictionary containing extinction model keyword arguments
        :type kwargs_extinction: `dict`
        :param kwargs_tracer_source: dictionary containing tracer source model keyword
        :return: prior
        :rtype: float
        """
        prior = 0.0

        # Limit the difference between pa_light and pa_mass for the deflector, where pa is the
        # position angle of the major axis
        if (
            "lens_option" in self.settings
            and "limit_mass_pa_from_light" in self.settings["lens_option"]
        ):
            max_mass_pa_difference = self.settings["lens_option"][
                "limit_mass_pa_from_light"
            ]

            if not isinstance(max_mass_pa_difference, (int, float)):
                raise ValueError(
                    "The value for limit_mass_pa_from_light should be a number!"
                )

            pa_mass = (
                ellipticity2phi_q(kwargs_lens[0]["e1"], kwargs_lens[0]["e2"])[0]
                * 180
                / np.pi
            )
            pa_light = (
                ellipticity2phi_q(
                    kwargs_lens_light[0]["e1"], kwargs_lens_light[0]["e2"]
                )[0]
                * 180
                / np.pi
            )

            diff = min(abs(pa_light - pa_mass), 180 - abs(pa_light - pa_mass))
            if diff > np.abs(max_mass_pa_difference):
                prior += -((diff - np.abs(max_mass_pa_difference)) ** 2) / 1e-3

        # Limit the difference between q_light and q_mass for the deflector, where q is the axis
        # ratio of the elliptical profile
        if (
            "lens_option" in self.settings
            and "limit_mass_q_from_light" in self.settings["lens_option"]
        ):
            max_mass_q_difference = self.settings["lens_option"][
                "limit_mass_q_from_light"
            ]

            if not isinstance(max_mass_q_difference, (int, float)):
                raise ValueError(
                    "The value for limit_mass_q_from_light should be a number!"
                )

            q_mass = ellipticity2phi_q(kwargs_lens[0]["e1"], kwargs_lens[0]["e2"])[1]
            q_light = ellipticity2phi_q(
                kwargs_lens_light[0]["e1"], kwargs_lens_light[0]["e2"]
            )[1]

            q_mass = ellipticity2phi_q(kwargs_lens[0]["e1"], kwargs_lens[0]["e2"])[1]
            q_light = ellipticity2phi_q(
                kwargs_lens_light[0]["e1"], kwargs_lens_light[0]["e2"]
            )[1]
            diff = q_light - q_mass
            if diff > max_mass_q_difference:
                prior += -((diff - max_mass_q_difference) ** 2) / 1e-4

        # Provide logarithmic_prior on the source light profile beta param
        if (
            "source_light_option" in self.settings
            and "shapelet_scale_logarithmic_prior"
            in self.settings["source_light_option"]
        ):
            if self.settings["source_light_option"]["shapelet_scale_logarithmic_prior"]:
                for i, model in enumerate(self.get_source_light_model_list()):
                    if model == "SHAPELETS":
                        beta = kwargs_source[i]["beta"]
                        prior += -np.log(beta)

        return prior

    @staticmethod
    def load_mask(mask_file_path):
        """Load mask from file.

        :param mask_file_path: path to the mask file
        :type mask_file_path: `str`
        :return: mask
        :rtype: `numpy.ndarray`
        """

    def get_masks(self):
        """Create masks.

        :return:
        :rtype:
        """
        if "mask" in self.settings and self.settings["mask"] is not None:
            if (
                "provided" in self.settings["mask"]
                and self.settings["mask"]["provided"]
            ):
                masks = []
                for n in range(self.number_of_bands):
                    masks.append(
                        self._file_system.load_mask(
                            self.settings["lens_name"], self.settings["band"][n]
                        )
                    )
            else:
                masks = []
                mask_options = deepcopy(self.settings["mask"])

                for n in range(self.number_of_bands):
                    band_name = self.settings["band"][n]

                    image_data = self.get_image_data(band_name)
                    coordinate_system = image_data.get_image_coordinate_system()
                    num_pixel = image_data.get_image_size()

                    # ra_at_xy_0 = mask_options["ra_at_xy_0"][n]
                    # dec_at_xy_0 = mask_options["dec_at_xy_0"][n]
                    # transform_pix2angle = np.array(
                    #     mask_options["transform_matrix"][n]
                    # )
                    # num_pixel = mask_options["size"][n]
                    # coords = Coordinates(
                    #     transform_pix2angle, ra_at_xy_0, dec_at_xy_0
                    # )

                    offset = mask_options["centroid_offset"][n]

                    x_coords, y_coords = coordinate_system.coordinate_grid(
                        num_pixel, num_pixel
                    )

                    if "radius" in mask_options:
                        radius = mask_options["radius"][n]
                        mask = mask_util.mask_azimuthal(
                            util.image2array(x_coords),
                            util.image2array(y_coords),
                            self.deflector_center_ra + offset[0],
                            self.deflector_center_dec + offset[1],
                            radius,
                        )
                    elif (
                        "a" in mask_options
                        and "b" in mask_options
                        and "angle" in mask_options
                    ):
                        a = mask_options["a"][n]
                        b = mask_options["b"][n]
                        angle = mask_options["angle"][n]
                        mask = mask_util.mask_ellipse(
                            util.image2array(x_coords),
                            util.image2array(y_coords),
                            self.deflector_center_ra + offset[0],
                            self.deflector_center_dec + offset[1],
                            a,
                            b,
                            angle,
                        )
                    else:
                        raise ValueError("Mask shape not properly specified!")

                    extra_masked_regions = []
                    try:
                        self.settings["mask"]["extra_regions"]
                    except (NameError, KeyError):
                        pass
                    else:
                        if self.settings["mask"]["extra_regions"] is not None:
                            for reg in self.settings["mask"]["extra_regions"][n]:
                                extra_masked_regions.append(
                                    1
                                    - mask_util.mask_azimuthal(
                                        util.image2array(x_coords),
                                        util.image2array(y_coords),
                                        self.deflector_center_ra + reg[0],
                                        self.deflector_center_dec + reg[1],
                                        reg[2],
                                    )
                                )

                    for extra_region in extra_masked_regions:
                        mask *= extra_region

                    # Mask Edge Pixels
                    if "mask_edge_pixels" in self.settings["mask"]:
                        border_length = self.settings["mask"]["mask_edge_pixels"][n]
                        if border_length > 0:
                            edge_mask = 0 * np.ones((num_pixel, num_pixel), dtype=int)

                            edge_mask[
                                border_length:-border_length,
                                border_length:-border_length,
                            ] = 1
                            edge_mask = (edge_mask.flatten()).tolist()
                        elif border_length == 0:
                            edge_mask = 1 * np.ones((num_pixel, num_pixel), dtype=int)
                            edge_mask = (edge_mask.flatten()).tolist()

                        mask *= edge_mask

                    # sanity check
                    mask[mask >= 1.0] = 1.0
                    mask[mask <= 0.0] = 0.0

                    masks.append(util.array2image(mask))

            return masks
        else:
            return None

    def get_kwargs_psf_iteration(self):
        """Create `kwargs_psf_iteration`.

        :return:
        :rtype:
        """
        if (
            "psf_iteration" in self.settings["fitting"]
            and self.settings["fitting"]["psf_iteration"]
        ):
            kwargs_psf_iteration = {
                "stacking_method": "median",
                "keep_psf_variance_map": True,
                "psf_symmetry": 4,
                "block_center_neighbour": 0.0,
                "num_iter": 50,
                "psf_iter_factor": 0.5,
            }

            if "psf_iteration_settings" in self.settings["fitting"]:
                for key in self.settings["fitting"]["psf_iteration_settings"].keys():
                    kwargs_psf_iteration[key] = self.settings["fitting"][
                        "psf_iteration_settings"
                    ][key]

            return kwargs_psf_iteration
        else:
            return {}

    def get_kwargs_numerics(self):
        """Create `kwargs_numerics`.

        :return:
        :rtype:
        """
        try:
            self.settings["kwargs_numerics"]["supersampling_factor"]
        except (KeyError, NameError, TypeError):
            supersampling_factor = [3] * self.number_of_bands
        else:
            supersampling_factor = deepcopy(
                self.settings["kwargs_numerics"]["supersampling_factor"]
            )

            if supersampling_factor is None:
                supersampling_factor = [3] * self.number_of_bands

        kwargs_numerics = []
        for n in range(self.number_of_bands):
            kwargs_numerics.append(
                {
                    "supersampling_factor": supersampling_factor[n],
                    "supersampling_convolution": False,
                    "supersampling_kernel_size": 3,
                    "flux_evaluate_indexes": None,
                    "point_source_supersampling_factor": 1,
                    "compute_mode": "regular",
                }
            )

        return kwargs_numerics

    @property
    def num_satellites(self):
        """Check if the system has satellites.

        :return:
        :rtype:
        """
        if "satellites" not in self.settings:
            return 0
        else:
            return len(self.settings["satellites"]["centroid_init"])

    def get_lens_model_list(self):
        """Return `lens_model_list`."""
        return self.get_lens_model_list_with_flags()[0]

    def get_lens_model_list_with_flags(self):
        """Return `lens_model_list` and `satellite_flags`.

        :return: lens_model_list, satellite_flags
        :rtype: list, list
        """
        lens_model_list = []
        satellite_flag = []

        if "lens" in self.settings["model"]:
            lens_model_list = [model for model in self.settings["model"]["lens"]]
            satellite_flag = [-1 for _ in range(len(lens_model_list))]

            if self.num_satellites > 0:
                if "is_elliptical" not in self.settings["satellites"]:
                    is_elliptical = [False] * self.num_satellites
                else:
                    is_elliptical = self.settings["satellites"]["is_elliptical"]
                for i, yes in enumerate(is_elliptical):
                    if yes:
                        lens_model_list.append("SIE")
                    else:
                        lens_model_list.append("SIS")
                    satellite_flag.append(i)

        return lens_model_list, satellite_flag

    def get_source_light_model_list(self):
        """Return `source_model_list`.

        :return:
        :rtype:
        """
        source_light_model_list = []

        if "source_light" in self.settings["model"]:
            # return self.settings["model"]["source_light"]
            for i in range(self.number_of_bands):
                source_light_model_list += self.settings["model"]["source_light"]

        return source_light_model_list

    def get_lens_light_model_list(self):
        """Return `lens_light_model_list`."""
        return self.get_lens_light_model_list_with_flags()[0]

    def get_lens_light_model_list_with_flags(self):
        """Return `lens_light_model_list` and `satellite_flags`.

        :return: lens_light_model_list, satellite_flags
        :rtype: list, list
        """
        lens_light_model_list = []
        satellite_flag = []

        if "lens_light" in self.settings["model"]:
            for i in range(self.number_of_bands):
                lens_light_model_list += [
                    model for model in self.settings["model"]["lens_light"]
                ]
                satellite_flag += [-1 for model in self.settings["model"]["lens_light"]]

            if self.num_satellites > 0:
                if "is_elliptical" not in self.settings["satellites"]:
                    is_elliptical = [False] * self.num_satellites
                else:
                    is_elliptical = self.settings["satellites"]["is_elliptical"]

                for i in range(self.number_of_bands):
                    for j, yes in enumerate(is_elliptical):
                        if yes:
                            lens_light_model_list.append("SERSIC_ELLIPSE")
                        else:
                            lens_light_model_list.append("SERSIC")
                        satellite_flag.append(j)

        return lens_light_model_list, satellite_flag

    def get_point_source_model_list(self):
        """Return `ps_model_list`.

        :return:
        :rtype:
        """
        if (
            "point_source" in self.settings["model"]
            and self.settings["model"]["point_source"] is not None
        ):
            return self.settings["model"]["point_source"]
        else:
            return []

    def get_index_list(self, light_type="lens_light"):
        """Create list with of index for the different light profiles (for multiple
        filters)"""
        index_list = []

        if light_type in self.settings["model"]:
            index_list = [[] for _ in range(self.number_of_bands)]
            counter = 0
            n = len(self.settings["model"][light_type])

            for num_band in range(self.number_of_bands):
                for _ in range(n):
                    index_list[num_band].append(counter)
                    counter += 1

            if light_type == "lens_light":
                for i in range(self.num_satellites):
                    for num_band in range(self.number_of_bands):
                        index_list[num_band].append(counter)
                        counter += 1

        return index_list

    def get_index_lens_light_model_list(self):
        """Create list with of index for the different lens light profile (for multiple
        filters)"""
        index_list = self.get_index_list("lens_light")

        return index_list

    def get_index_source_light_model_list(self):
        """Create list with of index for the different source light profiles (for
        multiple filters)"""
        return self.get_index_list("source_light")

    def get_lens_model_params(self):
        """Create `lens_params`.

        :return:
        :rtype:
        """
        lens_model_list, satellite_flags = self.get_lens_model_list_with_flags()

        fixed = []
        init = []
        sigma = []
        lower = []
        upper = []

        for i, model in enumerate(lens_model_list):
            if satellite_flags[i] == -1:
                # central deflector
                bound = self.deflector_centroid_bound
                center_x = self.deflector_center_ra
                center_y = self.deflector_center_dec
                try:
                    theta_E_init = self.settings["guess_params"]["lens"][0]["theta_E"]
                except (NameError, KeyError):
                    theta_E_init = 1.0
            else:
                # satellite
                bound = self.settings["satellites"]["centroid_bound"]
                center_x = self.settings["satellites"]["centroid_init"][
                    satellite_flags[i]
                ][0]
                center_y = self.settings["satellites"]["centroid_init"][
                    satellite_flags[i]
                ][1]
                theta_E_init = 0.1

            if model in ["SPEP", "PEMD", "EPL", "SIE"]:
                if model == "SIE":
                    fixed.append({"gamma": 2.0})
                else:
                    fixed.append({})
                init.append(
                    {
                        "center_x": center_x,
                        "center_y": center_y,
                        "e1": 0.0,
                        "e2": 0.0,
                        "gamma": 2.0,
                        "theta_E": theta_E_init,
                    }
                )
                sigma.append(
                    {
                        "theta_E": 0.1,
                        "e1": 0.01,
                        "e2": 0.01,
                        "gamma": 0.02,
                        "center_x": bound / 3,
                        "center_y": bound / 3,
                    }
                )

                lower.append(
                    {
                        "theta_E": 0.3,
                        "e1": -0.5,
                        "e2": -0.5,
                        "gamma": 1.3,
                        "center_x": center_x - bound,
                        "center_y": center_y - bound,
                    }
                )

                upper.append(
                    {
                        "theta_E": 3.0,
                        "e1": 0.5,
                        "e2": 0.5,
                        "gamma": 2.8,
                        "center_x": center_x + bound,
                        "center_y": center_y + bound,
                    }
                )
            elif model == "SHEAR_GAMMA_PSI":
                fixed.append({"ra_0": 0, "dec_0": 0})
                init.append({"gamma_ext": 0.05, "psi_ext": 0.0})
                sigma.append({"gamma_ext": 0.05, "psi_ext": np.pi / 90.0})
                lower.append({"gamma_ext": 0.0, "psi_ext": -np.pi})
                upper.append({"gamma_ext": 0.5, "psi_ext": np.pi})
            elif model == "SIS":
                fixed.append({})
                init.append(
                    {
                        "center_x": center_x,
                        "center_y": center_y,
                        "theta_E": theta_E_init,
                    }
                )
                sigma.append(
                    {
                        "theta_E": 0.1,
                        "center_x": bound / 3,
                        "center_y": bound / 3,
                    }
                )
                lower.append(
                    {
                        "theta_E": 0.0,
                        "center_x": center_x - bound,
                        "center_y": center_y - bound,
                    }
                )
                upper.append(
                    {
                        "theta_E": 1.0,
                        "center_x": center_x + bound,
                        "center_y": center_y + bound,
                    }
                )
            else:
                raise ValueError("{} not implemented as a lens " "model!".format(model))

        fixed = self.fill_in_fixed_from_settings("lens", fixed)

        params = [init, sigma, fixed, lower, upper]
        return params

    def get_lens_light_model_params(self):
        """Create `lens_light_params`.

        :return:
        :rtype:
        """
        lens_light_model_list, satellite_flags = (
            self.get_lens_light_model_list_with_flags()
        )

        fixed = []
        init = []
        sigma = []
        lower = []
        upper = []

        for i, model in enumerate(lens_light_model_list):
            if satellite_flags[i] == -1:
                bound = self.deflector_centroid_bound
                center_x = self.deflector_center_ra
                center_y = self.deflector_center_dec
            else:
                bound = self.settings["satellites"]["centroid_bound"]
                center_x = self.settings["satellites"]["centroid_init"][
                    satellite_flags[i]
                ][0]
                center_y = self.settings["satellites"]["centroid_init"][
                    satellite_flags[i]
                ][1]
            if "SERSIC" in model:
                if satellite_flags[i] == -1:
                    _fixed = {}
                else:
                    _fixed = {"n_sersic": 4.0}
                _init = {
                    "amp": 1.0,
                    "R_sersic": 0.2,
                    "center_x": center_x,
                    "center_y": center_y,
                    "n_sersic": 4.0,
                }
                _sigma = {
                    "center_x": np.max(self.pixel_size) / 10.0,
                    "center_y": np.max(self.pixel_size) / 10.0,
                    "R_sersic": 0.05,
                    "n_sersic": 0.5,
                }
                _lower = {
                    "n_sersic": 0.5,
                    "R_sersic": 0.1,
                    "center_x": center_x - bound,
                    "center_y": center_y - bound,
                }
                _upper = {
                    "n_sersic": 8.0,
                    "R_sersic": 5.0,
                    "center_x": center_x + bound,
                    "center_y": center_y + bound,
                }
                if "_ELLIPSE" in model:
                    _init.update({"e1": 0.0, "e2": 0.0})
                    _sigma.update({"e1": 0.05, "e2": 0.05})
                    _lower.update({"e1": -0.5, "e2": -0.5})
                    _upper.update({"e1": 0.5, "e2": 0.5})

                fixed.append(_fixed)
                init.append(_init)
                sigma.append(_sigma)
                lower.append(_lower)
                upper.append(_upper)
            else:
                raise ValueError(
                    "{} not implemented as a lens light" "model!".format(model)
                )

        fixed = self.fill_in_fixed_from_settings("lens_light", fixed)

        params = [init, sigma, fixed, lower, upper]
        return params

    def get_source_light_model_params(self):
        """Create `source_params`.

        :return:
        :rtype:
        """
        source_light_model_list = self.get_source_light_model_list()

        fixed = []
        init = []
        sigma = []
        lower = []
        upper = []

        shapelets_index = 0
        for i, model in enumerate(source_light_model_list):
            if model == "SERSIC_ELLIPSE":
                fixed.append({})

                init.append(
                    {
                        "amp": 1.0,
                        "R_sersic": 0.2,
                        "n_sersic": 1.0,
                        "center_x": 0.0,
                        "center_y": 0.0,
                        "e1": 0.0,
                        "e2": 0.0,
                    }
                )

                sigma.append(
                    {
                        "center_x": 0.5,
                        "center_y": 0.5,
                        "R_sersic": 0.01,
                        "n_sersic": 0.5,
                        "e1": 0.05,
                        "e2": 0.05,
                    }
                )

                lower.append(
                    {
                        "R_sersic": 0.04,
                        "n_sersic": 0.5,
                        "center_y": -2.0,
                        "center_x": -2.0,
                        "e1": -0.5,
                        "e2": -0.5,
                    }
                )

                upper.append(
                    {
                        "R_sersic": 0.5,
                        "n_sersic": 8.0,
                        "center_y": 2.0,
                        "center_x": 2.0,
                        "e1": 0.5,
                        "e2": 0.5,
                    }
                )
            elif model == "SHAPELETS":
                # If n_max is given as a single integer, convert it to a list
                if isinstance(self.settings["source_light_option"]["n_max"], int):
                    self.settings["source_light_option"]["n_max"] = [
                        self.settings["source_light_option"]["n_max"]
                        for _ in range(self.number_of_bands)
                    ]

                fixed.append(
                    {
                        "n_max": self.settings["source_light_option"]["n_max"][
                            shapelets_index
                        ]
                    }
                )
                init.append(
                    {
                        "center_x": 0.0,
                        "center_y": 0.0,
                        "beta": 0.10,
                        "n_max": self.settings["source_light_option"]["n_max"][
                            shapelets_index
                        ],
                    }
                )
                sigma.append(
                    {"center_x": 0.5, "center_y": 0.5, "beta": 0.010 / 10.0, "n_max": 2}
                )
                lower.append(
                    {"center_x": -1.2, "center_y": -1.2, "beta": 0.02, "n_max": -1}
                )
                upper.append(
                    {"center_x": 1.2, "center_y": 1.2, "beta": 0.20, "n_max": 55}
                )
                shapelets_index += 1
            else:
                raise ValueError(
                    "{} not implemented as a source light" "model!".format(model)
                )

        fixed = self.fill_in_fixed_from_settings("source_light", fixed)

        params = [init, sigma, fixed, lower, upper]
        return params

    def get_point_source_params(self):
        """Create `ps_params`.

        :return:
        :rtype:
        """
        point_source_model_list = self.get_point_source_model_list()

        fixed = []
        init = []
        sigma = []
        lower = []
        upper = []

        if len(point_source_model_list) > 0:
            fixed.append({})

            init.append(
                {
                    "ra_image": np.array(
                        self.settings["point_source_option"]["ra_init"]
                    ),
                    "dec_image": np.array(
                        self.settings["point_source_option"]["dec_init"]
                    ),
                }
            )

            num_point_sources = len(init[0]["ra_image"])
            sigma.append(
                {
                    "ra_image": np.max(self.pixel_size) * np.ones(num_point_sources),
                    "dec_image": np.max(self.pixel_size) * np.ones(num_point_sources),
                }
            )

            lower.append(
                {
                    "ra_image": init[0]["ra_image"]
                    - self.settings["point_source_option"]["bound"],
                    "dec_image": init[0]["dec_image"]
                    - self.settings["point_source_option"]["bound"],
                }
            )

            upper.append(
                {
                    "ra_image": init[0]["ra_image"]
                    + self.settings["point_source_option"]["bound"],
                    "dec_image": init[0]["dec_image"]
                    + self.settings["point_source_option"]["bound"],
                }
            )

        params = [init, sigma, fixed, lower, upper]
        return params

    def fill_in_fixed_from_settings(self, component, fixed_list):
        """Fill in fixed values from settings for lens, source light and lens light.

        :param component: name of component, 'lens', 'lens_light', or
            'source_light'
        :type component: `str`
        :param fixed_list: list of fixed params
        :type fixed_list: `list`
        :return:
        :rtype:
        """
        assert component in ["lens", "lens_light", "source_light"]
        option_str = component + "_option"

        try:
            self.settings[option_str]["fix"]
        except (NameError, KeyError):
            pass
        else:
            if self.settings[option_str]["fix"] is not None:
                for index, param_dict in self.settings[option_str]["fix"].items():
                    for key, value in param_dict.items():
                        # Propagting the fixed values in light profile to all bands
                        if component in ["lens_light", "source_light"]:
                            for n in range(self.number_of_bands):
                                num_profiles = len(self.settings["model"][component])
                                fixed_list[int(index) + n * num_profiles][key] = value
                        elif (
                            component == "lens"
                        ):  # for mass model that doesn't need duplication for multiple bands
                            fixed_list[int(index)][key] = value

        return fixed_list

    def get_kwargs_params(self):
        """Create `kwargs_params`.

        :return:
        :rtype:
        """
        kwargs_params = {
            "lens_model": self.get_lens_model_params(),
            "source_model": self.get_source_light_model_params(),
            "lens_light_model": self.get_lens_light_model_params(),
            "point_source_model": self.get_point_source_params(),
            # 'cosmography': []
        }

        return kwargs_params

    def get_psf_supersampled_factor(self):
        """Retrieve PSF supersampling factor if specified in the config file.

        :return: PSF supersampling factor
        :rtype: `float`
        """
        try:
            self.settings["psf_supersampled_factor"]
        except (NameError, KeyError):
            return 1
        else:
            return self.settings["psf_supersampled_factor"]



# -*- coding: utf-8 -*-
"""This module takes an image data and create semantic segmentation for it."""

import numpy as np
from tensorflow.keras.models import load_model
from scipy.ndimage import zoom

from .ai import AI


class Vision(AI):
    """This class takes an image data and create semantic segmentation for it."""

    def __init__(self, io_directory_path, source_type="quasar"):
        """

        :param data_file_path: path to a data file
        :type data_file_path: `str`
        """
        if source_type not in ["quasar", "galaxy"]:
            raise ValueError(
                f"Invalid source type: {source_type}. It should be either 'quasar' or 'galaxy'."
            )
        super().__init__(io_directory_path)

        self._source_type = source_type
        self.nn_model_path = self.file_system.get_trained_nn_model_file_path(
            source_type=source_type
        )
        self.nn_model = load_model(self.nn_model_path, compile=False)

    def create_segmentation_for_all_lenses(self, band_name):
        """Create semantic segmentation for all lenses.

        :param band_name: band name
        :type band_name: `str`
        """
        lens_list = self.file_system.get_lens_list()

        for lens_name in lens_list:
            self.create_segmentation_for_single_lens(lens_name, band_name)

        print(f"Semantic segmentation for {len(lens_list)} lenses has been created.")

    def create_segmentation_for_single_lens(self, lens_name, band_name):
        """Create semantic segmentation for a single lens.

        :param lens_name: lens name
        :type lens_name: `str`
        :param band_name: band name
        :type band_name: `str`
        """
        image_data = self.get_image_data(lens_name, band_name)
        image = image_data.get_image()

        segmentation = self.get_semantic_segmentation_from_nn(image)

        self.save_segmentation(lens_name, band_name, segmentation)

        return segmentation

    def save_segmentation(self, lens_name, band_name, segmentation):
        """Save the segmentation to a file.

        :param lens_name: lens name
        :type lens_name: `str`
        :param band_name: band name
        :type band_name: `str`
        :param segmentation: semantic segmentation
        :type segmentation: `numpy.ndarray`
        """
        self.file_system.save_semantic_segmentation(lens_name, band_name, segmentation)

    @staticmethod
    def resize_image(image):
        """Resize the image to (128, 128, 1).

        :param image: image data
        :type image: `numpy.ndarray`
        :return: resized image
        :rtype: `numpy.ndarray`
        """
        # Target shape for spatial dimensions
        target_shape = (128, 128)

        zoom_factors = [
            target_shape[0] / image.shape[0],
            target_shape[1] / image.shape[1],
        ]
        resampled_image = zoom(
            image, zoom_factors, order=3
        )  # order=3 for bicubic interpolation
        return resampled_image

    @staticmethod
    def resize_segmentation_to_original_size(segmentation, original_size):
        """Resize the prediction to the original size.

        :param segmentation: predicted segmentation from the NN
        :type segmentation: `numpy.ndarray`
        :param original_size: original size of the image
        :type original_size: int
        :return: resized prediction
        :rtype: `numpy.ndarray`
        """
        segmentation_shape = segmentation.shape
        segmentation_reshaped = np.zeros((original_size, original_size))

        for i in range(original_size):
            for j in range(original_size):
                segmentation_reshaped[i, j] = segmentation[
                    int(i / float(original_size) * segmentation_shape[0]),
                    int(j / float(original_size) * segmentation_shape[1]),
                ]

        return segmentation_reshaped

    def get_semantic_segmentation_from_nn(self, image):
        """Get semantic segmentation for the image from the trained neural network.

        :param image: image data
        :type image: `numpy.ndarray`
        :return: semantic segmentation
        :rtype: `numpy.ndarray`
        """
        resized_image = self.resize_image(image)
        image_input = np.expand_dims(resized_image, axis=0)

        # Get predictions from the model
        prediction = self.nn_model.predict(image_input)  # Shape: (1, 128, 128, 5)

        segmentation = np.argmax(prediction[0], axis=-1)  # Shape: (128, 128)

        # Resize the segmentation to the original size
        reshaped_segmentation = self.resize_segmentation_to_original_size(
            segmentation, image.shape[0]
        )

        if self._source_type == "galaxy":
            # Setting the satellite label to 4 to match with the case of quasar
            reshaped_segmentation[reshaped_segmentation == 3] = 4

        return reshaped_segmentation



# -*- coding: utf-8 -*-
"""This module has a class for maintaining the file system."""
__author__ = "ajshajib"

from pathlib import Path
import json
import numpy as np
import h5py
import gdown


class FileSystem(object):
    """This class contains the method to handle the file system and directory
    addresses."""

    def __init__(self, io_directory):
        """Initiates a FileSystem object with `io_directory` as root.

        :param io_directory: path to input/output directory
        :type io_directory: str
        """
        self._root_path = Path(io_directory)
        self.root = io_directory

    @staticmethod
    def path2str(path):
        """Converts a pathlib Path into string.

        :param path: path to a file or directory
        :type path: `Path`
        :return: path to a file or directory
        :rtype: `str`
        """
        return str(path.resolve())

    def get_lens_list_file_path(self):
        """Get the address for the lens_list.txt file.

        :return:
        :rtype:
        """
        return self.path2str(self._root_path / "lens_list.txt")

    def get_lens_list(self):
        """Get the list of lenses from lens_list.txt.

        :return:
        :rtype:
        """
        lens_list = []

        for line in open(self.get_lens_list_file_path(), "r"):
            if not line.startswith("#"):
                lens_list.append(line.rstrip("\n").rstrip("\r"))

        return lens_list

    def get_config_file_path(self, lens_name):
        """Get the file path to the config file for `lens_name`.

        :param lens_name: lens name
        :type lens_name: `str`
        :return: path to the config file
        :rtype: `str`
        """
        return self.path2str(self.get_settings_directory() / f"{lens_name}_config.yaml")

    def get_logs_directory(self):
        """Get directory for logs folder. If the directory doesn't exist, a folder is
        created.

        :return:
        :rtype:
        """
        logs_dir = self.path2str(self._root_path / "logs")

        # commenting out, as this directory needs to be created by user
        # if not os.path.isdir(logs_dir):
        #    os.mkdir(logs_dir)

        return logs_dir

    def get_settings_directory(self):
        """Get directory for settings folder. If the directory doesn't exist, a folder
        is created.

        :return:
        :rtype:
        """
        return self._root_path / "settings"

        # commenting out, as this directory needs to be created by user
        # if not os.path.isdir(settings_dir):
        #     os.mkdir(settings_dir)

    def get_outputs_directory(self):
        """Get directory for settings folder. If the directory doesn't exist, a folder
        is created.

        :return:
        :rtype:
        """
        outputs_dir = self.path2str(self._root_path / "outputs")

        # commenting out, as this directory needs to be created by user
        # if not os.path.isdir(outputs_dir):
        #    os.mkdir(outputs_dir)

        return outputs_dir

    def get_data_directory(self):
        """Get directory for data folder. If the directory doesn't exist, a folder is
        created.

        :return:
        :rtype:
        """
        data_dir = self.path2str(self._root_path / "data")

        # commenting out, as this directory needs to be created by user
        # if not os.path.isdir(data_dir):
        #     os.mkdir(data_dir)

        return data_dir

    def get_image_file_path(self, lens_name, band):
        """Get the file path for the imaging data for `lens_name`.

        :param lens_name: lens name
        :type lens_name: `str`
        :param band: band name
        :type band: `str`
        :return: file path
        :rtype: `str`
        """
        return self.path2str(
            Path(self.get_data_directory())
            / f"{lens_name}"
            / f"image_{lens_name}_{band}.h5"
        )

    def get_psf_file_path(self, lens_name, band):
        """Get the file path for the PSF data for `lens_name`.

        :param lens_name: lens name
        :type lens_name: `str`
        :param band: band name
        :type band: `str`
        :return: file path
        :rtype: `str`
        """
        return self.path2str(
            Path(self.get_data_directory())
            / f"{lens_name}"
            / f"psf_{lens_name}_{band}.h5"
        )

    def get_log_file_path(self, lens_name, model_id):
        """Get the file path for the PSF data for `lens_name`.

        :param lens_name: lens name
        :type lens_name: `str`
        :param model_id: identifier for run model
        :type model_id: `str`
        :return: file path
        :rtype: `str`
        """
        return (
            self.path2str(Path(self.get_logs_directory()))
            + f"/log_{lens_name}_{model_id}.txt"
        )

    def get_output_file_path(self, lens_name, model_id, file_type="json"):
        """Get the file path for the PSF data for `lens_name`.

        :param lens_name: lens name
        :type lens_name: `str`
        :param model_id: identifier for run model
        :type model_id: `str`
        :param file_type: type of file, options: 'json', 'h5'
        :type file_type: `str`
        :return: file path
        :rtype: `str`
        """
        return (
            self.path2str(Path(self.get_outputs_directory()))
            + f"/output_{lens_name}_{model_id}.{file_type}"
        )

    def save_output(self, lens_name, model_id, output, file_type="h5"):
        """Save output from fitting sequence.

        :param lens_name: name of the lens
        :type lens_name: `str`
        :param model_id: identifier for model run
        :type model_id: `str`
        :param output: output dictionary
        :type output: `dict`
        :param file_type: type of file to save, 'h5' or 'json'
        :type file_type: `str`
        :return: None
        :rtype:
        """
        if file_type == "h5":
            self.save_output_h5(lens_name, model_id, output)
        elif file_type == "json":
            self.save_output_json(lens_name, model_id, output)
        else:
            raise ValueError(f"File type {file_type} not recognized!")

    def save_output_json(self, lens_name, model_id, output):
        """Save output from fitting sequence.

        :param lens_name: name of the lens
        :type lens_name: `str`
        :param model_id: identifier for model run
        :type model_id: `str`
        :param output: output dictionary
        :type output: `dict`
        :return: None
        :rtype:
        """
        save_file = self.get_output_file_path(lens_name, model_id, file_type="json")
        with open(save_file, "w") as f:
            json.dump(self.encode_numpy_arrays(output), f, ensure_ascii=False, indent=4)

    def save_output_h5(self, lens_name, model_id, output):
        """Save output from fitting sequence to h5 format.

        :param lens_name: name of the lens
        :type lens_name: `str`
        :param model_id: identifier for model run
        :type model_id: `str`
        :param output: output dictionary
        :type output: `dict`
        :return: None
        :rtype:
        """
        save_file = self.get_output_file_path(lens_name, model_id, file_type="h5")

        with h5py.File(save_file, "w") as f:
            f.attrs["settings"] = json.dumps(
                self.encode_numpy_arrays(output["settings"]), ensure_ascii=False
            )
            f.attrs["kwargs_result"] = json.dumps(
                self.encode_numpy_arrays(output["kwargs_result"]), ensure_ascii=False
            )

            f.attrs["multi_band_list_out"] = json.dumps(
                self.encode_numpy_arrays(output["multi_band_list_out"]),
                ensure_ascii=False,
            )

            group = f.create_group("fit_output")
            for i, single_output in enumerate(output["fit_output"]):
                subgroup = group.create_group(f"{i}")
                subgroup.attrs["fitting_type"] = np.string_(single_output[0])

                if single_output[0] == "PSO":
                    subgroup.create_dataset("chi2", data=np.array(single_output[1][0]))
                    subgroup.create_dataset(
                        "position", data=np.array(single_output[1][1])
                    )
                    subgroup.create_dataset(
                        "velocity", data=np.array(single_output[1][2])
                    )
                    subgroup.create_dataset(
                        "param_list", data=np.array(single_output[2], dtype="S25")
                    )
                elif single_output[0] == "emcee":
                    subgroup.create_dataset(
                        "samples",
                        data=np.array(
                            single_output[1],
                        ),
                    )
                    subgroup.create_dataset(
                        "param_list", data=np.array(single_output[2], dtype="S25")
                    )
                    subgroup.create_dataset(
                        "log_likelihood",
                        data=np.array(
                            single_output[3],
                        ),
                    )
                else:
                    raise ValueError(
                        f"Fitting type {single_output[0]} not recognized for "
                        "saving output!"
                    )

    def load_output(self, lens_name, model_id, file_type="h5"):
        """Load from saved output file.

        :param lens_name: lens name
        :type lens_name: `str`
        :param model_id: model identifier provided at run initiation
        :type model_id: `str`
        :param file_type: type of file, 'h5' or 'json'
        :return: output dictionary
        :rtype: `dict`
        """
        if file_type == "h5":
            return self.load_output_h5(lens_name, model_id)
        elif file_type == "json":
            return self.load_output_json(lens_name, model_id)
        else:
            raise ValueError(f"File type {file_type} not recognized!")

    def load_output_json(self, lens_name, model_id):
        """Load from saved output file.

        :param lens_name: lens name
        :type lens_name: `str`
        :param model_id: model identifier provided at run initiation
        :type model_id: `str`
        :return: output dictionary
        :rtype: `dict`
        """
        load_file = self.get_output_file_path(lens_name, model_id, file_type="json")

        with open(load_file, "r") as f:
            output = json.load(f)

        return self.decode_numpy_arrays(output)

    def load_output_h5(self, lens_name, model_id):
        """Load from saved output file.

        :param lens_name: lens name
        :type lens_name: `str`
        :param model_id: model identifier provided at run initiation
        :type model_id: `str`
        :return: output dictionary
        :rtype: `dict`
        """
        load_file = self.get_output_file_path(lens_name, model_id, file_type="h5")

        with h5py.File(load_file, "r") as f:
            settings = self.decode_numpy_arrays(json.loads(str(f.attrs["settings"])))

            kwargs_result = self.decode_numpy_arrays(
                json.loads(str(f.attrs["kwargs_result"]))
            )

            multi_band_list_out = self.decode_numpy_arrays(
                json.loads(str(f.attrs["multi_band_list_out"]))
            )

            fit_output = []
            group = f["fit_output"]

            n = len(f["fit_output"].keys())
            for index in [f"{i}" for i in range(n)]:
                fitting_step = [
                    str(group[index].attrs["fitting_type"], encoding="utf-8")
                ]

                if fitting_step[0] == "PSO":
                    fitting_step.append(
                        [
                            group[index]["chi2"][:],
                            group[index]["position"][:],
                            group[index]["velocity"][:],
                        ]
                    )
                    fitting_step.append(
                        [
                            str(s, encoding="utf-8")
                            for s in group[index]["param_list"][:]
                        ]
                    )
                elif fitting_step[0] == "emcee":
                    fitting_step.append(group[index]["samples"][:])
                    fitting_step.append(
                        [
                            str(s, encoding="utf-8")
                            for s in group[index]["param_list"][:]
                        ]
                    )
                    fitting_step.append(group[index]["log_likelihood"][:])
                # else:
                #     raise ValueError('Fitting type {} not recognized for '
                #                      'loading output!'.format(fitting_step[0]
                #                      ))

                fit_output.append(fitting_step)

            output = {
                "settings": settings,
                "kwargs_result": kwargs_result,
                "fit_output": fit_output,
                "multi_band_list_out": multi_band_list_out,
            }

            return output

    @classmethod
    def encode_numpy_arrays(cls, obj):
        """Encode a list/dictionary containing numpy arrays through recursion for JSON
        serialization.

        :param obj: object
        :type obj:
        :return: object with `ndarray`s encoded as dictionaries
        :rtype:
        """
        if isinstance(obj, np.ndarray):
            return {"__ndarray__": obj.tolist(), "shape": obj.shape}
        elif isinstance(obj, list):
            encoded = []
            for element in obj:
                encoded.append(cls.encode_numpy_arrays(element))
            return encoded
        elif isinstance(obj, dict):
            encoded = {}
            for key, value in obj.items():
                encoded[key] = cls.encode_numpy_arrays(value)
            return encoded
        else:
            return obj

    @classmethod
    def decode_numpy_arrays(cls, obj):
        """Decode a list/dictionary containing encoded numpy arrays through recursion.

        :param obj: object with `ndarray`s encoded as dictionaries
        :type obj:
        :return: object with `ndarray`s as `numpy.ndarray`
        :rtype:
        """
        if isinstance(obj, dict):
            if "__ndarray__" in obj:
                return np.asarray(obj["__ndarray__"]).reshape(obj["shape"])
            else:
                decoded = {}
                for key, value in obj.items():
                    decoded[key] = cls.decode_numpy_arrays(value)
                return decoded
        elif isinstance(obj, list):
            decoded = []
            for element in obj:
                decoded.append(cls.decode_numpy_arrays(element))
            return decoded
        else:
            return obj

    def get_semantic_segmentation_file_path(self, lens_name, band):
        """Get the file path for the semantic segmentation data for `lens_name`.

        :param lens_name: lens name
        :type lens_name: `str`
        :param band: band name
        :type band: `str`
        :return: file path
        :rtype: `str`
        """
        return self.path2str(
            Path(self.get_outputs_directory())
            / f"semantic_segmentation_{lens_name}_{band}.npy"
        )

    def load_semantic_segmentation(self, lens_name, band):
        """Load semantic segmentation data from file.

        :param lens_name: lens name
        :type lens_name: `str`
        :param band: band name
        :type band: `str`
        :return: semantic segmentation array
        :rtype: `numpy.ndarray`
        """
        semantic_segmentation = np.load(
            self.get_semantic_segmentation_file_path(lens_name, band)
        )

        return semantic_segmentation

    def save_semantic_segmentation(self, lens_name, band, semantic_segmentation):
        """Save semantic segmentation data to file.

        :param lens_name: lens name
        :type lens_name: `str`
        :param band: band name
        :type band: `str`
        :param semantic_segmentation: semantic segmentation array
        :type semantic_segmentation: `numpy.ndarray`
        :return: None
        :rtype:
        """
        save_file = self.get_semantic_segmentation_file_path(lens_name, band)
        np.save(save_file, semantic_segmentation)

    def get_mask_file_path(self, lens_name, band):
        """Get the file path for the mask data for `lens_name`.

        :param lens_name: lens name
        :type lens_name: `str`
        :param band: band name
        :type band: `str`
        :return: file path
        :rtype: `str`
        """
        return self.path2str(
            self.get_settings_directory() / "masks" / f"mask_{lens_name}_{band}.npy"
        )

    def load_mask(self, lens_name, band):
        """Load mask data from file.

        :param lens_name: lens name
        :type lens_name: `str`
        :param band: band name
        :type band: `str`
        :return: mask array
        :rtype: `numpy.ndarray`
        """
        mask = np.load(self.get_mask_file_path(lens_name, band))

        return mask

    def save_mask(self, lens_name, band, mask):
        """Save mask data to file.

        :param lens_name: lens name
        :type lens_name: `str`
        :param band: band name
        :type band: `str`
        :param mask: mask array
        :type mask: `numpy.ndarray`
        :return: None
        :rtype:
        """
        save_file = self.get_mask_file_path(lens_name, band)
        np.save(save_file, mask)

    def get_trained_nn_model_file_path(self, source_type="galaxy"):
        """Get the file path for the trained model.

        :param lens_type: type of lens, 'galaxy' or 'quasar'
        :type lens_type: `str`
        :return: file path
        :rtype: `str`
        """
        assert source_type in ["galaxy", "quasar"]
        path = self.path2str(
            self._root_path
            / "trained_nn"
            / f"lensed_{source_type}_segmentation_model.h5"
        )

        # Check if the directory exists, create if not
        if not Path(path).parent.is_dir():
            Path(path).parent.mkdir()

        # Check if the file exists
        if not Path(path).is_file():
            # Download the model using gdown
            index = ["galaxy", "quasar"].index(source_type)
            file_id = [
                "1MAR2i5WlLlW_mAub3lbLLIlL6MPGXG8s",
                "1xO6Mniir3169H-7K5nThXR4lLvOUp6OQ",
            ][index]

            print("AI model not found in local storage. Downloading from the web...")
            gdown.download(
                id=file_id,
                output=path,
                quiet=False,
            )

        return path



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages
from pathlib import Path

this_directory = Path(__file__).parent
readme = (this_directory / "README.rst").read_text()

requirements = []

test_requirements = []

setup(
    author="Anowar J. Shajib",
    author_email="ajshajib@gmail.com",
    python_requires=">=3.9",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Users",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
    description="Automated pipeline for lens modeling based on lenstronomy",
    install_requires=requirements,
    license="BSD 3-Clause License",
    long_description=readme,
    long_description_content_type="text/x-rst",
    include_package_data=True,
    keywords="dolphin",
    name="dolphin",
    packages=find_packages(include=["dolphin", "dolphin.*"]),
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/ajshajib/dolphin",
    version="1.1.3",
    zip_safe=False,
)
